import pandas as pd
import numpy as np
import src.glucose360.preprocessing as pp
import configparser
import json

import plotly.express as px
from plotly.subplots import make_subplots
import plotly.graph_objects as go

import os

from src.glucose360.features import percent_time_in_range, percent_time_in_tight_range, mean, CV, GMI

dir_path = os.path.dirname(os.path.realpath(__file__))
config_path = os.path.join(dir_path, "config.ini")
config = configparser.ConfigParser()
config.read(config_path)
ID = config['variables']['id']
GLUCOSE = config['variables']['glucose']
TIME = config['variables']['time']
BEFORE = config['variables']['before']
AFTER = config['variables']['after']
TYPE = config['variables']['type']
DESCRIPTION = config['variables']['description']

def daily_plot_all(
    df: pd.DataFrame,
    events: pd.DataFrame = None,
    save: str = None,
    height: int = 2000
):
    """Graphs daily plots for all of the patients in the given DataFrame. 
    Also saves these plots as PDFs and HTMLs if passed a valid path.

    :param df: the DataFrame (following package guidelines) containing the CGM data to plot
    :type df: 'pandas.DataFrame'
    :param events: a DataFrame containing any events to mark on the daily plots, defaults to None
    :type events: 'pandas.DataFrame', optional
    :param save: path of the location where the saved PDF and HTML versions of the plots are saved, defaults to None
    :type save: str, optional 
    :param height: the height (in pixels) of the resulting plot(s), defaults to 2000
    :type height: int, optional
    """
    for id, data in df.groupby(ID):
        daily_plot(data, id, height, events, save)

def daily_plot(
   df: pd.DataFrame,
   id: str,
   height: int = 2000,
   events: pd.DataFrame = None,
   save: bool = False,
   app: bool = False
):
   """Graphs a daily (time-series) plot for the given patient within the given DataFrame. 
    Also saves this plot as both a PDF and HTML file if passed a valid path.

    :param df: the DataFrame (following package guidelines) containing the CGM data to plot
    :type df: 'pandas.DataFrame'
    :param id: the identification of the patient whose CGM data to plot within the given DataFrame
    :type id: str
    :param events: a DataFrame containing any events to mark on the daily plots, defaults to None
    :type events: 'pandas.DataFrame', optional
    :param save: path of the location where the saved PDF and HTML versions of the plot are saved, defaults to None
    :type save: str, optional 
    :param height: the height (in pixels) of the resulting plot, defaults to 2000
    :type height: int, optional
    :param app: boolean indicating whether to return the Plotly figure instead of rendering it (used mainly within the web application), defaults to False
    :type app: bool, optional
    :return: None if app is False, otherwise the Plotly figure
    :rtype: 'plotly.graph_objects.Figure' | None
   """
   show_events = not (events is None or events[events[ID] == id].empty)
   data = df.loc[id].copy()
   data[TIME] = pd.to_datetime(data[TIME])
   data["Day"] = data[TIME].dt.date

   days = data["Day"].unique().astype(str).tolist()
   offset = pd.Timedelta(hours=1, minutes=30)

   rendered_types = []
   if show_events:
      event_types = events[TYPE].unique()
      with open('event_colors.json') as colors_file:
         color_dict = json.load(colors_file)
         colors = list(color_dict.values())
         color_map = {event_type: colors[i] for i, event_type in enumerate(event_types)}

   fig = make_subplots(rows=len(days), cols=2 if show_events else 1, column_widths=[0.66,0.34] if show_events else None,
                        specs=[[{"type":"scatter"}, {"type":"table"}] for _ in range(len(days))] if show_events else None,
                        horizontal_spacing=0.01 if show_events else None,
                        vertical_spacing=0.05 if show_events else None)

   num_events = 0

   for idx, (day, dataset) in enumerate(data.groupby("Day"), start=1):
      fig.add_trace(go.Scatter(x=dataset[TIME],y=dataset[GLUCOSE],mode='lines+markers',name=str(day), showlegend=False), row=idx, col=1)
      fig.update_xaxes(range=[pd.Timestamp(day) - offset, pd.Timestamp(day) + pd.Timedelta(days=1) + offset], row=idx, col=1, tickfont_size=20, titlefont_size=35)
      fig.update_yaxes(title_text="Glucose Value (mg/dL)", tickfont_size=20, titlefont_size=35, row=idx, col=1)

      if show_events:
         day_events = events[(events[TIME].dt.date == day) & (events[ID] == id)].sort_values(TIME)
         num_events = max(num_events, day_events.shape[0])
         if not day_events.empty:
            table_body = [day_events[TIME].dt.time.astype(str).tolist(), day_events[DESCRIPTION].tolist()]
            fig.add_trace(
               go.Table(
                  columnwidth=[25,25],
                  header=dict(values = [["<b>Time</b>"], ["<b>Description</b>"]],font=dict(size=22)),
                  cells=dict(values=table_body,align=['left', 'left'],font=dict(size=20))
               ), row=idx, col=2
            )

            for _, row in day_events.drop_duplicates(subset=[TIME, TYPE]).iterrows():
               already_rendered = (row[TYPE] in rendered_types)
               if (not already_rendered): rendered_types.append(row[TYPE])
               fig.add_shape(go.layout.Shape(
                  type="line", yref="y domain",
                  x0=pd.to_datetime(row[TIME]), y0=0,
                  x1=pd.to_datetime(row[TIME]), y1=1,
                  line_color=color_map[row[TYPE]], line_dash="dash",
                  name=row[TYPE], legendgroup=row[TYPE], showlegend=(not already_rendered)), row=idx, col=1)

   fig.update_yaxes(range=[min(np.min(data[GLUCOSE]), 60) - 10, max(np.max(data[GLUCOSE]), 180) + 10])
   fig.update_layout(title=f"Daily Plot for {id}", height=height, showlegend=False, titlefont_size=40)

   if save: 
      path = os.path.join(save, f"{id}_daily_plot")
      fig.write_image(path + ".pdf", width=1500, height=height)
      fig.write_image(path + ".png", width=1500, height=height)
      fig.write_image(path + ".jpeg", width=1500, height=height)
      fig.write_html(path +".html")

   if app: return fig
   fig.show()

def event_plot_all(df: pd.DataFrame, id: str, events: pd.DataFrame, type: str, save: bool = False):
    """Graphs all event plots of a certain type for the given patient within the given DataFrame. 

    :param df: the DataFrame (following package guidelines) containing the CGM data to plot
    :type df: 'pandas.DataFrame'
    :param id: the identification of the patient whose CGM data to plot within the given DataFrame
    :type id: str
    :param events: a DataFrame containing events to be plotted, defaults to None
    :type events: 'pandas.DataFrame'
    :param type: the type of events to plot
    :type type: str
    """
    relevant_events = events[(events[ID] == id) & (events[TYPE] == type)]
    for index, event in relevant_events.iterrows():
        event_plot(df, id, event, relevant_events, save=save)

def event_plot(df: pd.DataFrame, id: str, event: pd.Series, events: pd.DataFrame = None, save: bool = False, app: bool = False):
   """Graphs an event plot for the given patient within the given DataFrame. 

    :param df: the DataFrame (following package guidelines) containing the CGM data to plot
    :type df: 'pandas.DataFrame'
    :param id: the identification of the patient whose CGM data to plot within the given DataFrame
    :type id: str
    :param event: the event to be displayed
    :type event: 'pandas.Series'
    :param events: a DataFrame containing any extra events to be marked within the event plot, defaults to None
    :type events: 'pandas.DataFrame', optional
    :param app: boolean indicating whether to return the Plotly figure instead of rendering it (used mainly within the web application), defaults to False
    :type app: bool, optional
    :return: None if app is False, otherwise the Plotly figure
    :rtype: 'plotly.graph_objects.Figure' | None
    """
   if event[ID] != id: raise Exception("Given event does not match the 'id' given.")
   data = df.loc[id].copy()
   event[TIME] = pd.to_datetime(event[TIME])
   before = event[TIME] - pd.Timedelta(minutes=event[BEFORE])
   after = event[TIME] + pd.Timedelta(minutes=event[AFTER])

   data["Day"] = data[TIME].dt.date
   subplot_figs = [go.Scatter(x=dataset[TIME],y=dataset[GLUCOSE],mode='lines+markers', name=str(day)) for day, dataset in data.groupby("Day")]
   fig = go.Figure(data=subplot_figs, layout=go.Layout(title=f"Event Plot for {id}", titlefont_size=40, legend=dict(font=dict(size=20))))

   event_data = events[events[ID] == id] if events is not None else pd.DataFrame()
   if not event_data.empty: create_event_lines(fig, event_data)

   fig.update_xaxes(type="date", range=[before, after], tickfont_size=20, titlefont_size=35)
   fig.update_yaxes(title_text="Glucose Value (mg/dL)", range=[85, 170], tickfont_size=20, titlefont_size=35)

   if save: 
      path = os.path.join(save, f"{id}_event_plot")
      fig.write_image(path + ".png", width=1500, height=1000)
      fig.write_html(path +".html")

   if app: return fig
   fig.show()

def create_event_lines(fig: go.Figure, events: pd.DataFrame):
   """Marks vertical lines within the given plotly Figure for the given events

   :param fig: the plotly figure to mark vertical lines in
   :type fig: plotly.graph_objects.Figure
   :param events: Pandas DataFrame containing vevents to mark
   :type events: pandas.DataFrame''
   """
   event_types = events[TYPE].unique()
   events[TIME] = pd.to_datetime(events[TIME])
   with open('event_colors.json') as colors_file:
      color_dict = json.load(colors_file)
      colors = list(color_dict.values())
      color_map = {event_type: colors[i] for i, event_type in enumerate(event_types)}

      rendered_types = []
      for event in events.itertuples():
         time = getattr(event, TIME)
         type = getattr(event, TYPE)
         already_rendered = (type in rendered_types) 
         if (not already_rendered): rendered_types.append(type)
         fig.add_vline(x=time, line_width=3, line_dash="dash", line_color=color_map[type], name=type, legendgroup=type, showlegend=(not already_rendered))

def weekly_plot_all(df: pd.DataFrame, save: str = None, height: int = 1000):
   """Graphs weekly plots for all of the patients within the given DataFrame. 
    Also saves these plots as PDF and HTML files if passed a valid path.

    :param df: the DataFrame (following package guidelines) containing the CGM data to plot
    :type df: 'pandas.DataFrame'
    :param save: path of the location where the saved PDF and HTML versions of the plots are saved, defaults to None
    :type save: str, optional 
    :param height: the height (in pixels) of the resulting plot(s), defaults to 1000
    :type height: int, optional
   """
   for id, data in df.groupby(ID):
      weekly_plot(data, id, height)

def weekly_plot(df: pd.DataFrame, id: str, save: str = None, height: int = 1000, app = False):
   """Graphs a weekly (time-series) plot for the given patient within the given DataFrame. 
    Also saves this plot as a PDF and HTML if passed a valid path.

    :param df: the DataFrame (following package guidelines) containing the CGM data to plot
    :type df: 'pandas.DataFrame'
    :param id: the identification of the patient whose CGM data to plot within the given DataFrame
    :type id: str
    :param save: path of the location where the saved PDF and HTML versions of the plots are saved, defaults to None
    :type save: str, optional 
    :param height: the height (in pixels) of the resulting plot, defaults to 1000
    :type height: int, optional
    :param app: boolean indicating whether to return the Plotly figure instead of rendering it (used mainly within the web application), defaults to False
    :type app: bool, optional
    :return: None if app is False, otherwise the Plotly figure
    :rtype: 'plotly.graph_objects.Figure' | None
   """
   data = df.loc[id].reset_index().copy()
   data.set_index(TIME, inplace=True)

   weekly_data = data.groupby(pd.Grouper(freq='W'))
   weekly_dfs = [group for _, group in weekly_data]
   
   fig = make_subplots(rows=len(weekly_dfs), cols=1)
   for week_index in range(len(weekly_dfs)):
      week = weekly_dfs[week_index].reset_index()
      fig.add_trace(
         go.Scatter(
               x=week[TIME],
               y=week[GLUCOSE],
               mode='lines+markers',
         ), row=(week_index+1), col=1
      )
      fig.update_yaxes(title_text="Glucose Value (mg/dL)", row=(week_index+1), col=1)
   
   if len(weekly_dfs) > 1:
      offset_before = pd.Timedelta(hours=10)
      offset_after = pd.Timedelta(hours=10)
      first_end = pd.Timestamp(weekly_dfs[1].reset_index()[TIME].dt.date.iloc[0])
      first_start = first_end - pd.Timedelta(weeks=1)
      last_start = pd.Timestamp(weekly_dfs[-1].reset_index()[TIME].dt.date.iloc[0])
      last_end = last_start + pd.Timedelta(weeks=1)
      fig.update_xaxes(range=[first_start - offset_before, first_end + offset_after], row=1, col=1)
      fig.update_xaxes(range=[last_start - offset_before, last_end + offset_after], row=len(weekly_dfs), col=1)
   fig.update_yaxes(range=[min(np.min(data[GLUCOSE]), 60) - 10, max(np.max(data[GLUCOSE]), 180) + 10], tickfont_size=20, titlefont_size=15)
   fig.update_xaxes(tickformat="%B %d, %Y <br> (%a)", tickfont_size=15, titlefont_size=30)

   fig.update_layout(title=f"Weekly Plot for {id}", titlefont_size=30, height=height, showlegend=False)

   if save: 
      path = os.path.join(save, f"{id}_weekly_plot")
      fig.write_image(path + ".pdf", width=1500, height=height)
      fig.write_image(path + ".png", width=1500, height=height)
      fig.write_image(path + ".jpeg", width=1500, height=height)
      fig.write_html(path +".html")

   if app: return fig
   fig.show()

def spaghetti_plot_all(df: pd.DataFrame, chunk_day: bool = False, save: str = None, height: int = 600):
    """Graphs spaghetti plots for all patients within the given DataFrame. 
    Also saves these plots as PDF and HTML files if passed a valid path.

    :param df: the DataFrame (following package guidelines) containing the CGM data to plot
    :type df: 'pandas.DataFrame'
    :param chunk_day: boolean indicating whether to create separate subplots based on whether the data occurred on a weekday or during the weekend
    :type chunk_day: bool, optional
    :param save: path of the location where the saved PDF and HTML versions of the plots are saved, defaults to None
    :type save: str, optional 
    :param height: the height (in pixels) of the resulting plot(s), defaults to 600
    :type height: int, optional
   """
    for id, data in df.groupby(ID):
        spaghetti_plot(df=data, id=id, chunk_day=chunk_day, save=save, height=height)

def spaghetti_plot(df: pd.DataFrame, id: str, chunk_day: bool = False, save: str = None, height: int = 600, app=False):
    """Graphs a spaghetti plot for the given patient within the given DataFrame. 
    Also saves this plot as both a PDF and HTML file if passed a valid path.

    :param df: the DataFrame (following package guidelines) containing the CGM data to plot
    :type df: 'pandas.DataFrame'
    :param id: the identification of the patient whose CGM data to plot within the given DataFrame
    :type id: str
    :param chunk_day: boolean indicating whether to create separate subplots based on whether the data occurred on a weekday or during the weekend
    :type chunk_day: bool, optional
    :param save: path of the location where the saved PDF and HTML versions of the plots are saved, defaults to None
    :type save: str, optional 
    :param height: the height (in pixels) of the resulting plot(s), defaults to 600
    :type height: int, optional
    :param app: boolean indicating whether to return the Plotly figure instead of rendering it (used mainly within the web application), defaults to False
    :type app: bool, optional
    :return: None if app is False, otherwise the Plotly figure
    :rtype: 'plotly.graph_objects.Figure' | None
   """
    data = df.loc[id].copy()
    data["Day"] = data[TIME].dt.date
    times = data[TIME] - data[TIME].dt.normalize()
    data["Time"] = (pd.to_datetime(["1/1/1970" for i in range(data[TIME].size)]) + times)
    data.sort_values(by=[TIME], inplace=True)

    fig = px.line(data, x="Time", y=GLUCOSE, color="Day", title=f"Spaghetti Plot for {id}", height=height, facet_col="Day Chunking" if chunk_day else None)
    fig.update_xaxes(tickformat="%H:%M:%S", title_text="Time", tickfont_size=20, titlefont_size=35) # shows only the times for the x-axis
    fig.update_yaxes(title_text="Glucose Value (mg/dL)", tickfont_size=20, titlefont_size=35, row=1, col=1)
    fig.update_annotations(font_size=35)
    fig.update_layout(title=dict(font=dict(size=40)), legend=dict(font=dict(size=30)))

    if save: 
      path = os.path.join(save, f"{id}_spaghetti_plot")
      fig.write_image(path + ".pdf", width=1500, height=height)
      fig.write_image(path + ".png", width=1500, height=height)
      fig.write_image(path + ".jpeg", width=1500, height=height)
      fig.write_html(path +".html")

    if app: return fig
    fig.show()

def AGP_plot_all(df: pd.DataFrame, height: int = 600, save: str = None):
    """Graphs AGP-report style plots for all of the patients within the given DataFrame. 
    Also saves these plots as both PDF and HTML files if passed a valid path.

    :param df: the DataFrame (following package guidelines) containing the CGM data to plot
    :type df: 'pandas.DataFrame'
    :param save: path of the location where the saved PDF and HTML versions of the plot are saved, defaults to None
    :type save: str, optional 
    :param height: the height (in pixels) of the resulting plot, defaults to 600
    :type height: int, optional
   """
    for id, data in df.groupby(ID):
        AGP_plot(data, id, height)


def AGP_plot(df: pd.DataFrame, id: str, save: str = None, height: int = 600, app=False):
    """Graphs AGP-report style plots for the given patient within the given DataFrame. 
    Also saves this plot as both a PDF and HTML file if passed a valid path.

    :param df: the DataFrame (following package guidelines) containing the CGM data to plot
    :type df: 'pandas.DataFrame'
    :param id: the identification of the patient whose CGM data to plot within the given DataFrame
    :type id: str
    :param save: path of the location where the saved PDF and HTML versions of the plot are saved, defaults to None
    :type save: str, optional 
    :param height: the height (in pixels) of the resulting plot, defaults to 600
    :type height: int, optional
    :param app: boolean indicating whether to return the Plotly figure instead of rendering it (used mainly within the web application), defaults to False
    :type app: bool, optional
    :return: None if app is False, otherwise the Plotly figure
    :rtype: 'plotly.graph_objects.Figure' | None
   """
    config.read('config.ini')
    interval = int(config["variables"]["interval"])
    if interval > 5:
         raise Exception("Data needs to have measurement intervals at most 5 minutes long")

    data = df.loc[id].copy()  
    data.reset_index(inplace=True)

    data[[TIME, GLUCOSE, ID]] = pp._resample_data(data[[TIME, GLUCOSE, ID]])
    times = data[TIME] - data[TIME].dt.normalize()
    # need to be in a DateTime format so plots can tell how to scale the x axis labels below
    data["Time"] = (pd.to_datetime(["1/1/1970" for i in range(data[TIME].size)]) + times)

    data.set_index("Time", inplace=True)

    agp_data = pd.DataFrame()
    for time, measurements in data.groupby("Time"):
      metrics = {
         "Time": time,
         "5th": measurements[GLUCOSE].quantile(0.05),
         "25th": measurements[GLUCOSE].quantile(0.25),
         "Median": measurements[GLUCOSE].median(),
         "75th": measurements[GLUCOSE].quantile(0.75),
         "95th": measurements[GLUCOSE].quantile(0.95),
      }
      agp_data = pd.concat([agp_data, pd.DataFrame.from_records([metrics])])

    agp_data.sort_values(by=["Time"], inplace=True)

    fig = go.Figure()
    fig.add_trace(go.Scatter(name="5th", x=agp_data["Time"], y=agp_data["5th"], line=dict(color="#869FCE")))
    fig.add_trace(go.Scatter(name="25th", x=agp_data["Time"], y=agp_data["25th"], fill="tonexty", fillcolor="#C9D4E9", line=dict(color="#97A8CB")))
    fig.add_trace(go.Scatter(name="Median",x=agp_data["Time"], y=agp_data["Median"], fill="tonexty", fillcolor="#97A8CB", line=dict(color="#183260")))
    fig.add_trace(go.Scatter(name="75th", x=agp_data["Time"], y=agp_data["75th"], fill="tonexty", fillcolor="#97A8CB", line=dict(color="#97A8CB")))
    fig.add_trace(go.Scatter(name="95th", x=agp_data["Time"], y=agp_data["95th"], fill="tonexty", fillcolor="#C9D4E9", line=dict(color="#869FCE")))

    fig.add_hline(y=70, line_color="lime")
    fig.add_hline(y=140, line_color="lime")
    fig.add_hline(y=180, line_color="green")
    fig.update_layout(title={"text": f"AGP Plot for {id}", "font": {"size":30}}, height=height, yaxis_range = [35,405], legend=dict(font=dict(size=30)))
    fig.update_xaxes(tickformat="%H:%M:%S", title_text="Time", tickfont_size=20, titlefont_size=25) # shows only the times for the x-axis
    fig.update_yaxes(title_text="Glucose Value (mg/dL)", tickfont_size=20, titlefont_size=25)

    if app: return fig
    fig.show()

def AGP_report(df: pd.DataFrame, id: str, path: str = None):
   """Creates an AGP-report for the given patient within the given DataFrame. 
    Also saves this plot as a PDF file if passed a valid path.

    :param df: the DataFrame (following package guidelines) containing the CGM data to report on
    :type df: 'pandas.DataFrame'
    :param id: the identification of the patient whose CGM data to report on
    :type id: str
    :param path: path of the location where the saved PDF version of the plot is saved, defaults to None
    :type path: str, optional 
    :return: the AGP-report in string form if path is False, otherwise None
    :rtype: str | None
   """
   fig = make_subplots(rows = 1, cols = 2, specs=[[{"type": "table"}, {"type": "bar"}]])
   
   patient_data = df.loc[id]
   TIR = {"< 54 mg/dL": percent_time_in_range(patient_data, 0, 53),
                       "54 - 69 mg/dL": percent_time_in_range(patient_data, 54, 69),
                       "70 - 140 mg/dL": percent_time_in_tight_range(patient_data),
                       "141 - 180 mg/dL": percent_time_in_range(patient_data, 141, 180),
                       "181 - 250 mg/dL": percent_time_in_range(patient_data, 181, 250),
                       "> 250 mg/dL": percent_time_in_range(patient_data, 251, 400)}
   
   COLORS = {"< 54 mg/dL": "rgba(151,34,35,255)",
                       "54 - 69 mg/dL": "rgba(229,43,23,255)",
                       "70 - 140 mg/dL": "#00ff00",
                       "141 - 180 mg/dL": "rgba(82,173,79,255)",
                       "181 - 250 mg/dL": "rgba(250,192,3,255)",
                       "> 250 mg/dL": "rgba(241,136,64,255)"}
   
   for key, value in TIR.items():
      fig.add_trace(go.Bar(name=key, x=["TIR"], y=[value], marker=dict(color=COLORS[key]), text=[round(value, 2)], textposition="inside"), row=1, col=2)
   fig.update_layout(barmode='stack', height=600, font=dict(size=20))
   
   ave_glucose = mean(patient_data)
   gmi = GMI(patient_data)
   cv = CV(patient_data)

   days = patient_data[TIME].dt.date
   table_body = [["Test Patient ID:", f"{len(days.unique())} Days:", "Average Glucose (mg/dL):", "Glucose Management Indicator:", "Glucose Variability:"], 
                 [id, f"{days.iloc[0]} to {days.iloc[-1]}", str(round(ave_glucose, 2)), str(round(gmi, 2)), str(round(cv, 2))]]
   fig.add_trace(go.Table(cells=dict(values=table_body,align=['left', 'center'],font=dict(size=25),height=50)), row=1, col=1)
   
   agp = AGP_plot(df, id, app=True); 
   weekly = weekly_plot(df, id, height=500, app=True); 

   header_html = fig.to_html(full_html=False, include_plotlyjs="cdn")
   agp_html = agp.to_html(full_html=False, include_plotlyjs=False)
   weekly_html = weekly.to_html(full_html=False, include_plotlyjs=False)

   html_template = f"""
   <!DOCTYPE html>
   <html>
   <head>
      <title>AGP Report for {id}</title>
   </head>
   <body>
      <h1> AGP Report: Continuous Glucose Monitoring </h1>
      {header_html}
      {agp_html}
      {weekly_html}
   </body>
   </html>
   """

   if path:
      with open(os.path.join(path, f"{id}_AGP_Report.html"), "w") as f:
         f.write(html_template)

   return html_template


import os, glob
import pandas as pd
import configparser
import zipfile, tempfile

dir_path = os.path.dirname(os.path.realpath(__file__))
config_path = os.path.join(dir_path, "config.ini")
config = configparser.ConfigParser()
config.read(config_path)
ID = config['variables']['id']
GLUCOSE = config['variables']['glucose']
TIME = config['variables']['time']

# globals for glucose values to replace "Low" and "High" with in the CGM data
LOW = 40
HIGH = 400

def import_data(
   path: str,
   name: str = None,
   sensor: str = "dexcom",
   id_template: str = None,
   glucose: str = None,
   time: str = None,
   interval: int = 5,
   max_gap: int = 45,
   output = print
) -> pd.DataFrame:
    """Returns a Multiindexed Pandas DataFrame containing all of the csv data found at the given path.
    The path can lead to a directory, .zip file, or a .csv file. The returned DataFrame holds columns 
    for timestamps and glucose values, and is indexed by patient identifications

    :param path: the path of the directory/zip/csv to be parsed through
    :type path: str
    :param sensor: the CGM device model used (either dexcom, freestyle libre pro, or freestyle libre 2 / freestyle libre 3), defaults to 'dexcom'
    :type sensor: str, optional
    :param id_template: regex dictating how to parse each CSV file's name for the proper patient identification, defaults to None
    :type id_template: str, optional
    :param glucose: the name of the column containing the glucose values in the .csv files (if different than the default for the CGM sensor being used), defaults to None
    :type glucose: str, optional
    :param time: the name of the column containing the timestamps in the .csv files (if different than the default for the CGM sensor being used), defaults to None
    :type time: str, optional
    :param interval: the resampling interval (in minutes) that the data should follow, defaults to 5
    :type interval: int, optional
    :param max_gap: the maximum amount of minutes a gap in the data can be interpolated, defaults to 45 
       (filling in a gap with a longer duration would be considered extrapolation)
    :type max_gap: int, optional
    :return: A Pandas DataFrame containing the preprocessed data found at the given path. This DataFrame holds columns for timestamps, glucose values, weekday/weekend chunking, and waking/sleeping time chunking.
    :rtype: pandas.DataFrame 

    :Example:
    >>> path_to_data = "datasets/patient_data.csv"
    >>> df = import_data(path_to_data)
    """
    # update the config with the resampling interval the user chose
    updated_config = config['variables'] 
    updated_config['interval'] = str(interval)
    config["variables"] = updated_config
    with open('config.ini', 'w') as configfile:
      config.write(configfile)

    # get file extension of where the given path points
    ext = os.path.splitext(path)[1]

    # path leads to directory
    if ext == "":
       if not os.path.isdir(path):
          raise ValueError("Directory does not exist")
       else:
          return _import_directory(path, sensor, id_template, glucose, time, interval, max_gap, output)
    
    # check if path leads to .zip or .csv
    if ext.lower() in [".csv", ".zip"]:
       if not os.path.isfile(path):
          raise ValueError("File does not exist")
    else:
       raise ValueError("Invalid file type")
   
    # path leads to .csv
    if ext.lower() == ".csv":
       return _import_csv(path, sensor, id_template, glucose, time, interval, max_gap)

    # otherwise has to be a .zip file
    with zipfile.ZipFile(path, 'r') as zip_ref:
      # create a temporary directory to pull from
      with tempfile.TemporaryDirectory() as temp_dir:
         zip_ref.extractall(temp_dir)
         dir = name or path.split("/")[-1].split(".")[0]
         return _import_directory((temp_dir + "/" + dir), sensor, id_template, glucose, time, interval, max_gap, output)
    
def _import_directory(
    path: str,
    sensor: str = "dexcom",
    id_template: str = None,
    glucose: str = None,
    time: str = None,
    interval: int = 5,
    max_gap: int = 45,   
    output = print
) -> pd.DataFrame:
    """Returns a Multiindexed Pandas DataFrame containing all of the csv data found at the given path.
    The path must lead to a directory containing .csv files. The returned DataFrame holds columns 
    for timestamps and glucose values, and is indexed by patient identifications

    :param path: the path of the directory to be parsed through
    :type path: str
    :param sensor: the CGM device model used (either dexcom, freestyle libre pro, or freestyle libre 2 / freestyle libre 3), defaults to 'dexcom'
    :type sensor: str, optional
    :param id_template: regex dictating how to parse each CSV file's name for the proper patient identification, defaults to None
    :type id_template: str, optional
    :param glucose: the name of the column containing the glucose values in the .csv files (if different than the default for the CGM sensor being used), defaults to None
    :type glucose: str, optional
    :param time: the name of the column containing the timestamps in the .csv files (if different than the default for the CGM sensor being used), defaults to None
    :type time: str, optional
    :param interval: the resampling interval (in minutes) that the data should follow, defaults to 5
    :type interval: int, optional
    :param max_gap: the maximum amount of minutes a gap in the data can be interpolated, defaults to 45 
       (filling in a gap with a longer duration would be considered extrapolation)
    :type max_gap: int, optional
    :return: A Pandas DataFrame containing the preprocessed data found at the given path. This DataFrame holds columns for timestamps, glucose values, weekday/weekend chunking, and waking/sleeping time chunking.
    :rtype: pandas.DataFrame 
    """
    csv_files = glob.glob(path + "/*.csv")
    num_files = len(csv_files)

    if num_files == 0:
       raise Exception("No CSV files found.")
    
    output(f"{num_files} .csv files were found in the specified directory.")
    
    data: list[pd.DataFrame] = []
    num_valid_files = num_files
    for file in csv_files:
       try:
          data.append(_import_csv(file, sensor, id_template, glucose, time, interval, max_gap))
       except:
          num_valid_files -= 1
   
    output(f"{num_valid_files} .csv files were successfully imported.")

    if len(data) == 0: raise Exception("CSV files found, but none were valid.")  
    df = pd.concat(data)

    output(f"{df.index.unique().size} sections were found in the imported data.")

    return df

def _import_csv(
    path: str,
    sensor: str = "dexcom", 
    id_template: str = None,
    glucose: str = None,
    time: str = None,
    interval: int = 5, 
    max_gap: int = 45
) -> pd.DataFrame:
    """Returns a Multiindexed Pandas DataFrame containing all of the csv data found at the given path.
    The path must lead to a .csv file. The returned DataFrame holds columns 
    for timestamps and glucose values, and is indexed by patient identifications

    :param path: the path of the csv file to be parsed through
    :type path: str
    :param sensor: the CGM device model used (either 'dexcom', 'freestyle libre pro', 'freestyle libre 2', 'freestyle libre 3', or 'columns'), defaults to 'dexcom'
    :type sensor: str, optional
    :param id_template: regex dictating how to parse the CSV file's name for the proper patient identification, 
       or the name of the patient identification column if using the 'columns' sensor, defaults to None
    :type id_template: str, optional
    :param glucose: the name of the column containing the glucose values in the .csv files (if different than the default for the CGM sensor being used), defaults to None
    :type glucose: str, optional
    :param time: the name of the column containing the timestamps in the .csv files (if different than the default for the CGM sensor being used), defaults to None
    :type time: str, optional
    :param interval: the resampling interval (in minutes) that the data should follow, defaults to 5
    :type interval: int, optional
    :param max_gap: the maximum amount of minutes a gap in the data can be interpolated, defaults to 45 
       (filling in a gap with a longer duration would be considered extrapolation)
    :type max_gap: int, optional
    :return: A Pandas DataFrame containing the preprocessed data found at the given path. This DataFrame holds columns for timestamps, glucose values, weekday/weekend chunking, and waking/sleeping time chunking.
    :rtype: pandas.DataFrame 
    """
    data = pd.DataFrame()
    if sensor == "dexcom":
       data = _import_csv_dexcom(path, id_template, glucose, time)
    elif sensor == "freestyle libre 2" or sensor == "freestyle libre 3":
       data = _import_csv_freestyle_libre_23(path, id_template, glucose, time)
    elif sensor == "freestyle libre pro":
       data = _import_csv_freestyle_libre_pro(path, id_template, glucose, time)
    elif sensor == "columns":
       data = _import_csv_columns(path, id_template, glucose, time)
   
    preprocessed_data = preprocess_data(data, interval, max_gap)
    return preprocessed_data

def _import_csv_columns(
   path: str,
   id_col: str = None,
   glucose_col: str = None,
   time_col: str = None,
) -> pd.DataFrame:
   """Returns a Pandas DataFrame containing all of the csv data found at the given path.
   The path must lead to a .csv file with three columns (identification, timestamp, and glucose value) containing CGM data. The returned DataFrame holds columns 
   for timestamps, glucose values, and the patient identification

   :param path: the path of the csv file to be parsed through
   :type path: str
   :param id_col: the name of the column containing the patient identification(s), defaults to None
   :type id_col: str, optional
   :param glucose_col: the name of the column containing the glucose values in the .csv files (if different than the default for the CGM sensor being used), defaults to None
   :type glucose_col: str, optional
   :param time_col: the name of the column containing the timestamps in the .csv files (if different than the default for the CGM sensor being used), defaults to None
   :type time_col: str, optional
   :return: A Pandas DataFrame containing the raw data found at the given path. This DataFrame holds columns for timestamps, glucose values, and the patient identification.
   :rtype: pandas.DataFrame 
   """
   df = pd.read_csv(path)
   glucose = glucose_col or "Glucose Value (mg/dL)"
   time = time_col or "Timestamp (YYYY-MM-DDThh:mm:ss)"
   id = id_col or "ID"

   df.rename(columns={glucose: GLUCOSE, time: TIME, id: ID}, inplace=True)
   return df

def _import_csv_dexcom(
    path: str,
    id_template: str = None,
    glucose_col: str = None,
    time_col: str = None,
) -> pd.DataFrame:
    """Returns a Pandas DataFrame containing all of the Dexcom csv data found at the given path.
    The path must lead to a .csv file containing CGM data from a Dexcom device. The returned DataFrame holds columns 
    for timestamps, glucose values, and the patient identification

    :param path: the path of the Dexcom csv file to be parsed through
    :type path: str
    :param id_template: regex dictating how to parse the CSV file's name for the proper patient identification, defaults to None
    :type id_template: str, optional
    :param glucose_col: the name of the column containing the glucose values in the .csv files (if different than the default for the CGM sensor being used), defaults to None
    :type glucose_col: str, optional
    :param time_col: the name of the column containing the timestamps in the .csv files (if different than the default for the CGM sensor being used), defaults to None
    :type time_col: str, optional
    :return: A Pandas DataFrame containing the raw data found at the given path. This DataFrame holds columns for timestamps, glucose values, and the patient identification.
    :rtype: pandas.DataFrame 
    """
    df = pd.read_csv(path)
    glucose = glucose_col or "Glucose Value (mg/dL)"
    time = time_col or "Timestamp (YYYY-MM-DDThh:mm:ss)"
    id = _retrieve_id_dexcom(path.split("/")[-1], df, id_template)

    df.rename(columns={glucose: GLUCOSE, time: TIME}, inplace=True)
    df[ID] = id
    return df

def _import_csv_freestyle_libre_23(
   path: str,
   id_template: str = None,
   glucose_col: str = None,
   time_col: str = None,  
) -> pd.DataFrame:
   """Returns a Pandas DataFrame containing all of the FreeStyle Libre 2 or 3 csv data found at the given path.
    The path must lead to a .csv file containing CGM data from FreeStyle Libre 2 or FreeStyle Libre 3 devices. The returned DataFrame holds columns 
    for timestamps, glucose values, and the patient identification

    :param path: the path of the FreeStyle Libre 2 or 3 csv file to be parsed through
    :type path: str
    :param id_template: regex dictating how to parse the CSV file's name for the proper patient identification, defaults to None
    :type id_template: str, optional
    :param glucose_col: the name of the column containing the glucose values in the .csv files (if different than the default for the CGM sensor being used), defaults to None
    :type glucose_col: str, optional
    :param time_col: the name of the column containing the timestamps in the .csv files (if different than the default for the CGM sensor being used), defaults to None
    :type time_col: str, optional
    :return: A Pandas DataFrame containing the raw data found at the given path. This DataFrame holds columns for timestamps, glucose values, and the patient identification.
    :rtype: pandas.DataFrame 
    """
   glucose = glucose_col or "Historic Glucose mg/dL"
   time = time_col or "Device Timestamp"
   return _import_csv_freestyle_libre(path, id_template, glucose, time)

def _import_csv_freestyle_libre_pro(
    path: str,
    id_template: str = None,
    glucose_col: str = None,
    time_col: str = None,
) -> pd.DataFrame:
    """Returns a Pandas DataFrame containing all of the FreeStyle Libre Pro csv data found at the given path.
    The path must lead to a .csv file containing CGM data from a FreeStyle Libre Pro device. The returned DataFrame holds columns 
    for timestamps, glucose values, and the patient identification

    :param path: the path of the FreeStyle Libre Pro csv file to be parsed through
    :type path: str
    :param id_template: regex dictating how to parse the CSV file's name for the proper patient identification, defaults to None
    :type id_template: str, optional
    :param glucose_col: the name of the column containing the glucose values in the .csv files (if different than the default for the CGM sensor being used), defaults to None
    :type glucose_col: str, optional
    :param time_col: the name of the column containing the timestamps in the .csv files (if different than the default for the CGM sensor being used), defaults to None
    :type time_col: str, optional
    :return: A Pandas DataFrame containing the raw data found at the given path. This DataFrame holds columns for timestamps, glucose values, and the patient identification.
    :rtype: pandas.DataFrame 
    """
    glucose = glucose_col or "Historic Glucose(mg/dL)"
    time = time_col or "Meter Timestamp"
    return _import_csv_freestyle_libre(path, id_template, glucose, time)

def _import_csv_freestyle_libre(
   path: str,
   id_template: str,
   glucose_col: str,
   time_col: str
) -> pd.DataFrame:
   """Returns a Pandas DataFrame containing all of the FreeStyle Libre csv data found at the given path.
    The path must lead to a .csv file containing CGM data from FreeStyle Libre 2/3/Pro devices. The returned DataFrame holds columns 
    for timestamps, glucose values, and the patient identification. 

    :param path: the path of the FreeStyle Libre csv file to be parsed through
    :type path: str
    :param id_template: regex dictating how to parse the CSV file's name for the proper patient identification
    :type id_template: str
    :param glucose_col: the name of the column containing the glucose values in the .csv files (if different than the default for the CGM sensor being used)
    :type glucose_col: str
    :param time_col: the name of the column containing the timestamps in the .csv files (if different than the default for the CGM sensor being used)
    :type time_col: str
    :return: A Pandas DataFrame containing the raw data found at the given path. This DataFrame holds columns for timestamps, glucose values, and the patient identification.
    :rtype: pandas.DataFrame 
    """
   id = pd.read_csv(path, nrows=1)["Patient report"].iloc[0] if not id_template else _id_from_filename(path.split("/")[-1], id_template)
   df = pd.read_csv(path, skiprows=2)

   df.rename(columns={glucose_col: GLUCOSE, time_col: TIME}, inplace=True)
   df[ID] = id
   return df

def _retrieve_id_dexcom(name: str, df: pd.DataFrame, id_template: str = None) -> str:
   """Returns the appropriate identification for the given raw Dexcom CGM data based on the given template.
   If the template is None, the identification will be pulled from the patient information fields from within the CSV.
   Otherwise, the filename will be parsed accordingly.

   :param name: the name of the file to parse for an identification
   :type name: str
   :param df: a Pandas DataFrame containing the raw data from a Dexcom CSV file
   :type df: pandas.DataFrame
   :param id_template: regex indicating how to parse the filename for the identification, defaults to None
   :type id_template: str, optional
   :return: the proper identification for the raw data in the given dataframe
   :rtype: str
   """
   if id_template and "first" not in id_template and "last" not in id_template and "patient_identifier" not in id_template:
      # need to parse file name for id
      return _id_from_filename(name, id_template)

   # use Dexcom fields for id instead
   first = df["Patient Info"].iloc[0]
   last = df["Patient Info"].iloc[1]
   patient_identifier = df["Patient Info"].iloc[2]
   id = df["Patient Info"].iloc[0] + df["Patient Info"].iloc[1] 
   if id_template: id = id_template.format(first=first, last=last, patient_identifier=patient_identifier)
   return id

def _id_from_filename(name: str, id_template: str):
   """Parses the given filename for an identification using a regex template.

   :param name: the filename to parse for an identification
   :type name: str
   :param id_template: regex indicating how to parse the filename for the identification
   :type id_template: str
   :return: the identification from the filename
   :rtype: str
   """
   import re
   pattern = re.compile(fr"{id_template}")
   match = pattern.search(name)
   if match is None:
      raise Exception("The RegEx ID template passed does not match the file name.")
   id = str(match.group("id"))
   try: 
      section = str(match.group("section"))
      id += f" ({section})"
   except:
      print(f"'Section' not defined for patient {id}.")
   return id

def preprocess_data(
   df: pd.DataFrame,
   interval: int = 5,
   max_gap: int = 45
) -> pd.DataFrame:
   """Returns a Pandas DataFrame containing the preprocessed CGM data within the given dataframe.
   As part of the preprocessing phase, the data will be converted into the proper data types, resampled, interpolated, chunked, and
   indexed by identification (alongside all 'Low's and 'High's being replaced and all edge null values being dropped)
   
   :param df: the Pandas DataFrame containing the CGM data to preprocess
   :type df: pandas.DataFrame
   :param interval: the resampling interval (in minutes) the CGM data should follow, defaults to 5
   :type interval: int, optional
   :param max_gap: the maximum duration (in minutes) of a gap in the data that should be interpolated, defaults to 45
   :type max_gap: int, optional
   :return: A Pandas DataFrame containing the preprocessed CGM data. This DataFrame is indexed by identification and holds columns for
      timestamps, glucose values, day chunking, and time chunking.
   :rtype: pandas.DataFrame 

   :Example:
   >>> # 'df' is a Pandas DataFrame already containing your CGM data, with columns for glucose values, timestamps, and identification
   >>> preprocessed_df = preprocess_data(df)
   """
   df = df.dropna(subset=[GLUCOSE])
   df = df.replace("Low", LOW)
   df = df.replace("High", HIGH)
   df.reset_index(drop=True, inplace=True)

   df[TIME] = pd.to_datetime(df[TIME])
   df[GLUCOSE] = pd.to_numeric(df[GLUCOSE])

   df = df[[TIME, GLUCOSE, ID]].copy()
   df = _resample_data(df, interval, max_gap)
   df = df.loc[df[GLUCOSE].first_valid_index():df[GLUCOSE].last_valid_index()]
   df = _chunk_day(_chunk_time(df))
   df.set_index(ID, inplace=True)
   return df

def _resample_data(df: pd.DataFrame, minutes: int = 5, max_gap: int = 45) -> pd.DataFrame:
    """Resamples and (if needed) interpolates the given default-indexed DataFrame.
    Used mostly to preprocess the data in the csv files being imported in _import_csv().

    :param df: the DataFrame to be resampled and interpolated
    :type df: pandas.DataFrame
    :param minutes: the length of the interval to be resampled into (in minutes), defaults to 5
    :type minutes: int
    :param max_gap: the maximum duration (in minutes) of gaps that should be interpolated, defaults to 45
    :type max_gap: int
    :return: A Pandas DataFrame containing the resampled and interpolated. This DataFrame holds columns for timestamps, glucose values, and the patient identification.
    :rtype: pandas.DataFrame
    """
    id = df.at[0, ID]

    # Sort the DataFrame by datetime
    resampled_df = df.sort_values(by=[TIME])
    resampled_df = resampled_df.set_index(TIME)

    interval = str(minutes) + "T"
    # generate the times that match the frequency
    resampled_df = resampled_df.asfreq(interval)
    # add in the original points that don't match the frequency (just for linear time-based interpolation)
    resampled_df.reset_index(inplace=True)
    resampled_df = (pd.concat([resampled_df, df])).drop_duplicates(subset=[TIME])
    resampled_df.sort_values(by=[TIME], inplace=True)

    # interpolate the missing values
    resampled_df.set_index(TIME, inplace=True)
    resampled_df = _interpolate_data(resampled_df, max_gap)
    
    # now that the values have been interpolated, remove the points that don't match the frequency
    resampled_df = resampled_df.asfreq(interval)
    resampled_df[ID] = id # resampled data points might have empty ID values
    resampled_df.reset_index(inplace=True)

    return resampled_df

def _interpolate_data(df: pd.DataFrame, max_gap: int) -> pd.DataFrame:
    """Only linearly interpolates NaN glucose values for time gaps that are less than the given number of minutes.
    Used mainly in preprocessing for csv files that are being imported in _import_csv().
    
    :param df: the Pandas DataFrame containing the CGM data to interpolate
    :type df: pandas.DataFrame
    :param max_gap: the maximum minute length of gaps that should be interpolated
    :type max_gap: int
    :return: a Pandas DataFrame with interpolated CGM data
    :rtype: pandas.DataFrame
    """
    config.read('config.ini')
    interval = int(config["variables"]["interval"])

    # based heavily on https://stackoverflow.com/questions/67128364/how-to-limit-pandas-interpolation-when-there-is-more-nan-than-the-limit

    s = df[GLUCOSE].notnull()
    s = s.ne(s.shift()).cumsum()

    m = df.groupby([s, df[GLUCOSE].isnull()])[GLUCOSE].transform('size').where(df[GLUCOSE].isnull())
    interpolated_df = df.interpolate(method="time", limit_area="inside").mask(m >= int(max_gap / interval))

    return interpolated_df

def _chunk_time(df: pd.DataFrame) -> pd.DataFrame:
    """Adds a new column to the given DataFrame specifying whether the values occur during a waking or sleeping period

    :param df: the Pandas DataFrame to add the new column to (must contain a column for timestamps)
    :type df: pandas.DataFrame
    :return: the Pandas DataFrame with the added column for time chunking
    :rtype: pandas.DataFrame
    """
    times = df[TIME] - df[TIME].dt.normalize()
    is_waking = (times >= pd.Timedelta(hours=8)) & (times <= pd.Timedelta(hours=22))
    df["Time Chunking"] = is_waking.replace({True: "Waking", False: "Sleeping"})
    return df

def _chunk_day(df: pd.DataFrame) -> pd.DataFrame:
    """Adds a new column to the given DataFrame specifying whether the values occur during a weekday or the weekend

    :param df: the Pandas DataFrame to add the new column to (must contain a column for timestamps)
    :type df: pandas.DataFrame
    :return: the Pandas DataFrame with the added column for day chunking
    :rtype: pandas.DataFrame
    """
    is_weekend = df[TIME].dt.dayofweek > 4
    df["Day Chunking"] = is_weekend.replace({True: "Weekend", False: "Weekday"})
    return df

def segment_data(path: str, df: pd.DataFrame) -> pd.DataFrame:
    """
    Splits patients' data into multiple segments based on a given .csv file containing ID's and DateTimes.
    
    :param path: path of the .csv file containing identifications and timestamps indicating where to split the given DataFrame
    :type path: str
    :param df: the DataFrame to split based on the given .csv file
    :type df: pandas.DataFrame
    :return: a Pandas DataFrame with the data split accordingly
    :rtype: pandas.DataFrame
    """
    # Read the segments CSV file
    segments = pd.read_csv(path)
    segments[TIME] = pd.to_datetime(segments[TIME])
    
    # Sort segments by TIME
    segments.sort_values(['ID', TIME], inplace=True)
    
    # Create a copy of the original dataframe to avoid modifying it directly
    df_copy = df.copy()
    df_copy[TIME] = pd.to_datetime(df_copy[TIME])
    df_copy = df_copy.reset_index()
    
    # Initialize a Segment column in the original dataframe
    df_copy['Segment'] = 0
    
    # Use a dictionary to keep track of segment counters
    segment_counters = {id: 1 for id in segments['ID'].unique()}
    
    # Iterate over each row in the segments dataframe
    for _, segment_row in segments.iterrows():
        id = segment_row['ID']
        time = segment_row[TIME]
        
        # Create a mask for rows before the current segment time
        mask = (df_copy['ID'] == id) & (df_copy[TIME] < time) & (df_copy['Segment'] == 0)
        
        # Update the segment counter for those rows
        df_copy.loc[mask, 'Segment'] = segment_counters[id]
        
        # Increment the segment counter
        segment_counters[id] += 1
    
    # Assign the segment counter to rows after the last date
    for id in segment_counters.keys():
        mask = (df_copy['ID'] == id) & (df_copy['Segment'] == 0)
        df_copy.loc[mask, 'Segment'] = segment_counters[id]
    
    # Combine ID and Segment to form the new ID
    df_copy['ID'] = df_copy['ID'].astype(str) + '_' + df_copy['Segment'].astype(str)
    
    # Drop the Segment column as it's no longer needed
    df_copy.drop(columns=['Segment'], inplace=True)
    
    return df_copy


import pandas as pd
import numpy as np
import configparser
from multiprocessing import Pool
import os
from scipy.integrate import trapezoid

dir_path = os.path.dirname(os.path.realpath(__file__))
config_path = os.path.join(dir_path, "config.ini")
config = configparser.ConfigParser()
config.read(config_path)
ID = config['variables']['id']
GLUCOSE = config['variables']['glucose']
TIME = config['variables']['time']

"""
All of the metric-calculating functions are designed for DataFrames that contain only one patient's data.
For example, if 'df' is the outputted DataFrame from 'import_data()', 'LBGI(df)' would not be accurate.
Instead, do 'LBGI(df.loc[PATIENT_ID])'.
"""

def mean(df: pd.DataFrame) -> float:
   """Calculates the mean glucose level for the given CGM trace

   :param df: a Pandas DataFrame containing the preprocessed CGM data to calculate the mean glucose for
   :type df: 'pandas.DataFrame'
   :return: the mean glucose level of the given CGM trace
   :rtype: float
   """
   return df[GLUCOSE].mean()

def summary_stats(df: pd.DataFrame) -> list[float]:
    """Calculates summary statistics (minimum, first quartile, median, third quartile, and maximum) for the given CGM trace

    :param df: a Pandas DataFrame containing the preprocessed CGM data to calculate the five-point summary for
    :type df: 'pandas.DataFrame'
    :return: a list containing the five-point summary for the given CGM trace
    :rtype: list[float]
    """
    min = df[GLUCOSE].min()
    first = df[GLUCOSE].quantile(0.25)
    median = df[GLUCOSE].median()
    third = df[GLUCOSE].quantile(0.75)
    max = df[GLUCOSE].max()

    return [min, first, median, third, max]

def SD(df: pd.DataFrame) -> float:
    """Calculates the standard deviation for the given CGM trace

    :param df: a Pandas DataFrame containing the preprocessed CGM data to calculate the standard deviation for
    :type df: 'pandas.DataFrame'
    :return: the standard deviation of the given CGM trace
    :rtype: float
    """
    return df[GLUCOSE].std()

def CV(df: pd.DataFrame) -> float:
   """Calculates the coefficient of variation (CV) for the given CGM trace

   :param df: a Pandas DataFrame containing the preprocessed CGM data to calculate the CV for
   :type df: 'pandas.DataFrame'
   :return: the CV of the given CGM trace
   :rtype: float
   """
   return SD(df) / mean(df) * 100

def eA1c(df: pd.DataFrame) -> float:
    """Calculates the estimated A1c (eA1c) for the given CGM trace

    :param df: a Pandas DataFrame containing the preprocessed CGM data to calculate the eA1c for
    :type df: 'pandas.DataFrame'
    :return: the eA1c of the given CGM trace
    :rtype: float
    """
    return (46.7 + mean(df)) / 28.7

def GMI(df: pd.DataFrame) -> float:
    """Calculates the Glucose Management Indicator (GMI) for the given CGM trace

    :param df: a Pandas DataFrame containing the preprocessed CGM data to calculate the GMI for
    :type df: 'pandas.DataFrame'
    :return: the GMI of the given CGM trace
    :rtype: float
    """
    return (0.02392 * mean(df)) + 3.31

def percent_time_in_range(df: pd.DataFrame, low: int = 70, high: int = 180) -> float:
   """Returns the percent of total time the given CGM trace's glucose levels were between the given lower and upper bounds (inclusive)
   
   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :param low: the lower bound of the acceptable glucose values, defaults to 70
   :type low: int, optional
   :param high: the upper bound of the acceptable glucose values, defaults to 180
   :type high: int, optional
   :return: the percentage of total time the glucose levels within the given CGM trace were between the given bounds (inclusive)
   :rtype: float
   """
   valid_df = df.dropna(subset=[GLUCOSE])
   in_range_df = valid_df[(valid_df[GLUCOSE] <= high) & (valid_df[GLUCOSE] >= low)]
   time_in_range = len(in_range_df)
   total_time = len(valid_df)
   return (100 * time_in_range / total_time) if total_time > 0 else np.nan

def percent_time_in_tight_range(df: pd.DataFrame):
   """Returns the percent of total time the given CGM trace's glucose levels were within 70-140 mg/dL (inclusive)

   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :return: the percentage of total time the glucose levels within the given CGM trace were within 70-140 mg/dL (inclusive)
   :rtype: float
   """
   return percent_time_in_range(df, low = 70, high = 140)

def percent_time_above_range(df: pd.DataFrame, limit: int = 180) -> float:
   """Returns the percent of total time the given CGM trace's glucose levels were above a given threshold (inclusive)
   
   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :param limit: the threshold for calculating the percent time above, defaults to 180
   :type limit: int, optional
   :return: the percentage of total time the glucose levels within the given CGM trace were above the given threshold (inclusive)
   :rtype: float
   """
   return percent_time_in_range(df, low = limit, high = 400)

def percent_time_below_range(df: pd.DataFrame, limit: int = 70) -> float:
   """Returns the percent of total time the given CGM trace's glucose levels were below a given threshold (inclusive)
   
   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :param limit: the threshold for calculating the percent time below, defaults to 70
   :type limit: int, optional
   :return: the percentage of total time the glucose levels within the given CGM trace were below the given threshold (inclusive)
   :rtype: float
   """
   return percent_time_in_range(df, low = 40, high = limit)

def percent_time_in_hypoglycemia(df: pd.DataFrame) -> float:
   """Returns the percent of total time the given CGM trace's glucose levels were within literature-defined
   ranges that indicate hypoglycemia
   
   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :return: the percentage of total time the glucose levels within the given CGM trace were in ranges indicating hypoglycemia (< 70 mg/dL)
   :rtype: float
   """
   return percent_time_below_range(df, 70)

def percent_time_in_level_1_hypoglycemia(df: pd.DataFrame) -> float:
   """Returns the percent of total time the given CGM trace's glucose levels were within literature-defined
   ranges that indicate level 1 hypoglycemia
   
   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :return: the percentage of total time the glucose levels within the given CGM trace were in ranges indicating level 1 hypoglycemia (54-70 mg/dL)
   :rtype: float
   """
   return percent_time_in_range(df, 54, 69)

def percent_time_in_level_2_hypoglycemia(df: pd.DataFrame) -> float:
   """Returns the percent of total time the given CGM trace's glucose levels were within literature-defined
   ranges that indicate level 2 hypoglycemia
   
   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :return: the percentage of total time the glucose levels within the given CGM trace were in ranges indicating level 2 hypoglycemia (< 54 mg/dL)
   :rtype: float
   """
   return percent_time_below_range(df, 53)

def percent_time_in_hyperglycemia(df: pd.DataFrame) -> float:
   """Returns the percent of total time the given CGM trace's glucose levels were within literature-defined
   ranges that indicate hyperglycemia
   
   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :return: the percentage of total time the glucose levels within the given CGM trace were in ranges indicating hyperglycemia (> 180 mg/dL)
   :rtype: float
   """
   return percent_time_above_range(df, 180)

def percent_time_in_level_0_hyperglycemia(df: pd.DataFrame) -> float:
   """Returns the percent of total time the given CGM trace's glucose levels were within 
   ranges that indicate level 0 hyperglycemia
   
   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :return: the percentage of total time the glucose levels within the given CGM trace were in ranges indicating level 0 hyperglycemia (140-180 mg/dL)
   :rtype: float
   """
   return percent_time_in_range(df, 140, 180)

def percent_time_in_level_1_hyperglycemia(df: pd.DataFrame) -> float:
   """Returns the percent of total time the given CGM trace's glucose levels were within literature-defined
   ranges that indicate level 1 hyperglycemia
   
   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :return: the percentage of total time the glucose levels within the given CGM trace were in ranges indicating level 1 hyperglycemia (180-250 mg/dL)
   :rtype: float
   """
   return percent_time_in_range(df, 181, 249)

def percent_time_in_level_2_hyperglycemia(df: pd.DataFrame) -> float:
   """Returns the percent of total time the given CGM trace's glucose levels were within literature-defined
   ranges that indicate level 2 hyperglycemia
   
   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :return: the percentage of total time the glucose levels within the given CGM trace were in ranges indicating level 2 hyperglycemia (> 250 mg/dL)
   :rtype: float
   """
   return percent_time_above_range(df, 250)

def ADRR(df: pd.DataFrame) -> float:
   """Calculates the Average Daily Risk Range (ADRR) for the given CGM trace.

   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :return: the ADRR for the given CGM trace
   :rtype: float
   """
   data = df.copy()

   # Convert time to date
   data['date'] = pd.to_datetime(data[TIME]).dt.date
   data = data.dropna(subset=[GLUCOSE])
   data['bgi'] = (np.log(data[GLUCOSE]) ** 1.084) - 5.381
   data['right'] = 22.7 * np.maximum(data['bgi'], 0) ** 2
   data['left'] = 22.7 * np.minimum(data['bgi'], 0) ** 2

   adrr = data.groupby(['date']).apply(lambda df: np.max(df['left']) + np.max(df['right'])).mean()
   return adrr

def BG_formula(ser: pd.Series) -> pd.Series:
    """Calculates the Average Daily Risk Range (ADRR) for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the ADRR for the given CGM trace
    :rtype: float
    """
    return 1.509 * (np.power(np.log(ser), 1.084) - 5.381)

def LBGI(df: pd.DataFrame) -> float:
    """Calculates the Low Blood Glucose Index (LBGI) for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the LBGI for the given CGM trace
    :rtype: float
    """
    BG = np.minimum(0, BG_formula(df[GLUCOSE]))
    return np.mean(10 * (BG ** 2))

def HBGI(df: pd.DataFrame) -> float:
    """Calculates the High Blood Glucose Index (HBGI) for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the HBGI for the given CGM trace
    :rtype: float
    """
    BG = np.maximum(0, BG_formula(df[GLUCOSE]))
    return np.mean(10 * (BG ** 2))

def COGI(df: pd.DataFrame) -> float:
    """Calculates the Continuous Glucose Monitoring Index (COGI) for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the COGI for the given CGM trace
    :rtype: float
    """
    tir = percent_time_in_range(df)
    tir_score = 0.5 * tir

    tbr = percent_time_in_range(df, 0, 70)
    tbr_score = 0.35 * ((1 - (np.minimum(tbr, 15) / 15)) * 100)

    sd = SD(df)
    sd_score = 100
    if sd >= 108:
        sd_score = 0
    elif sd > 18:
        sd_score = (1 - ((sd-18) / 90)) * 100
    sd_score *= 0.15
    
    COGI = tir_score + tbr_score + sd_score
    return COGI

def GRADE_formula(df: pd.DataFrame) -> pd.DataFrame:
    """Transforms each glucose value within the given CGM trace as needed to help calculate 
    the Glycaemic Risk Assessment Diabetes Equation (GRADE).

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the LBGI for the given CGM trace
    :rtype: float
    """
    df_GRADE = pd.DataFrame()
    df_GRADE[GLUCOSE] = df[GLUCOSE].copy()
    df_GRADE["GRADE"] = ((np.log10(np.log10(df[GLUCOSE] / 18)) + 0.16) ** 2) * 425
    return df_GRADE

def GRADE_eugly(df: pd.DataFrame) -> float:
    """Calculates the Glycaemic Risk Assessment Diabetes Equation (GRADE) for 
    solely the glucose values in target range (70-140 mg/dL) within the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the euglycemic GRADE for the given CGM trace
    :rtype: float
    """
    df_GRADE = GRADE_formula(df)
    return np.sum(df_GRADE[(df_GRADE[GLUCOSE] >= 70) & (df_GRADE[GLUCOSE] <= 140)]["GRADE"]) / np.sum(df_GRADE["GRADE"]) * 100

def GRADE_hypo(df: pd.DataFrame) -> float:
    """Calculates the Glycaemic Risk Assessment Diabetes Equation (GRADE) for 
    solely the glucose values in hypoglycemic range (<70 mg/dL) within the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the hypoglycemic GRADE for the given CGM trace
    :rtype: float
    """
    df_GRADE = GRADE_formula(df)
    return np.sum(df_GRADE[df_GRADE[GLUCOSE] < 70]["GRADE"]) / np.sum(df_GRADE["GRADE"]) * 100

def GRADE_hyper(df: pd.DataFrame) -> float:
    """Calculates the Glycaemic Risk Assessment Diabetes Equation (GRADE) for 
    solely the glucose values in hyperglycemic range (>140 mg/dL) within the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the hyperglycemic GRADE for the given CGM trace
    :rtype: float
    """
    df_GRADE = GRADE_formula(df)
    return np.sum(df_GRADE[df_GRADE[GLUCOSE] > 140]["GRADE"]) / np.sum(df_GRADE["GRADE"]) * 100

def GRADE(df: pd.DataFrame) -> float:
    """Calculates the Glycaemic Risk Assessment Diabetes Equation (GRADE) for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the GRADE for the given CGM trace
    :rtype: float
    """
    df_GRADE = GRADE_formula(df)
    return df_GRADE["GRADE"].mean()

def GRI(df: pd.DataFrame) -> float:
    """Calculates the Glycemia Risk Index (GRI) for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the GRI for the given CGM trace
    :rtype: float
    """
    vlow = percent_time_in_range(df, 0, 53)
    low = percent_time_in_range(df, 54, 69)
    high = percent_time_in_range(df, 181, 250)
    vhigh = percent_time_in_range(df, 251, 500) 

    return min((3 * vlow) + (2.4 * low) + (0.8 * high) + (1.6 * vhigh), 100)

def GVP(df: pd.DataFrame) -> float:
    """Calculates the Glucose Variability Percentage (GVP) for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the GVP for the given CGM trace
    :rtype: float
    """
    copy_df = df.dropna(subset=["Glucose"])
    delta_x = pd.Series(5, index=np.arange(copy_df.shape[0]), dtype="float", name='orders')
    delta_y = copy_df.reset_index()["Glucose"].diff()
    L = np.sum(np.sqrt((delta_x ** 2) + (delta_y ** 2)))
    L_0 = np.sum(delta_x)
    return L / L_0

def hyper_index(df: pd.DataFrame, limit: int = 140, a: float = 1.1, c: float = 30) -> float:
    """Calculates the Hyperglycemia Index for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :param limit: upper limit of target range (above which would hyperglycemia), defaults to 140 mg/dL
    :type limit: int, optional
    :param a: exponent utilized for Hyperglycemia Index calculation, defaults to 1.1
    :type a: float, optional
    :param c: constant to help scale Hyperglycemia Index the same as other metrics (e.g. LBGI, HBGI, and GRADE), defaults to 30
    :type c: float, optional
    :return: the Hyperglycemia Index for the given CGM trace
    :rtype: float
    """
    BG = df[GLUCOSE].dropna()
    return np.sum(np.power(BG[BG > limit] - limit, a)) / (BG.size * c)

def hypo_index(df: pd.DataFrame, limit: int = 80, b: float = 2, d: float = 30) -> float:
    """Calculates the Hypoglycemia Index for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :param limit: lower limit of target range (above which would hypoglycemia), defaults to 80 mg/dL
    :type limit: int, optional
    :param b: exponent utilized for Hypoglycemia Index calculation, defaults to 2
    :type b: float, optional
    :param d: constant to help scale Hypoglycemia Index the same as other metrics (e.g. LBGI, HBGI, and GRADE), defaults to 30
    :type d: float, optional
    :return: the Hypoglycemia Index for the given CGM trace
    :rtype: float
    """
    BG = df[GLUCOSE].dropna()
    return np.sum(np.power(limit - BG[BG < limit], b)) / (BG.size * d)

def IGC(df: pd.DataFrame) -> float:
    """Calculates the Index of Glycemic Control (IGC) for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the IGC for the given CGM trace
    :rtype: float
    """
    return hyper_index(df) + hypo_index(df)

def j_index(df: pd.DataFrame) -> float:
    """Calculates the J-Index for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the J-Index for the given CGM trace
    :rtype: float
    """
    return 0.001 * ((mean(df) + SD(df)) ** 2)

def CONGA(df: pd.DataFrame, n: int = 24) -> float:
    """Calculates the Continuous Overall Net Glycemic Action (CONGA) for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :param n: the difference in time (in hours) between observations used to calculate CONGA, defaults to 24
    :type n: int, optional
    :return: the CONGA for the given CGM trace
    :rtype: float
    """
    config.read('config.ini')
    interval = int(config["variables"]["interval"])
    period = n * (60 / interval)
    return np.std(df[GLUCOSE].diff(periods=period))

# lag is in days
def MODD(df: pd.DataFrame, lag: int = 1) -> float:
    """Calculates the Mean Difference Between Glucose Values Obtained at the Same Time of Day (MODD) for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :param lag: the difference in time (in days) between observations used to calculate MODD, defaults to 1
    :type lag: int, optional
    :return: the MODD for the given CGM trace
    :rtype: float
    """
    config.read('config.ini')
    interval = int(config["variables"]["interval"])
    period = lag * 24 * (60 / interval)
    
    return np.mean(np.abs(df[GLUCOSE].diff(periods=period)))

def mean_absolute_differences(df: pd.DataFrame) -> float:
    """Calculates the Mean Absolute Differences for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the mean absolute differences for the given CGM trace
    :rtype: float
    """
    return np.mean(np.abs(df[GLUCOSE].diff()))

def median_absolute_deviation(df: pd.DataFrame, constant: float = 1.4826) -> float:
    """Calculates the Median Absolute Deviation for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :param constant: factor to multiply median absolute deviation by, defaults to 1.4826
    :type constant: float, optional
    :return: the median absolute deviation for the given CGM trace
    :rtype: float
    """
    return constant * np.nanmedian(np.abs(df[GLUCOSE] - np.nanmedian(df[GLUCOSE])))

def MAG(df: pd.DataFrame) -> float:
   """Calculates the Mean Absolute Glucose (MAG) for the given CGM trace.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the MAG for the given CGM trace
    :rtype: float
    """
   df.dropna(subset=[GLUCOSE], inplace=True)
   data = df[(df[TIME].dt.minute == (df[TIME].dt.minute).iloc[0]) & (df[TIME].dt.second == (df[TIME].dt.second).iloc[0])][GLUCOSE]
   return np.sum(data.diff().abs()) / data.size

def MAGE(df: pd.DataFrame, short_ma: int = 5, long_ma: int = 32, max_gap: int = 180) -> float:
   """Calculates the Mean Amplitude of Glycemic Excursions (MAGE) for the given CGM trace.

   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :param short_ma: number of data points utilized to calculate short moving average values, defaults to 5
   :type short_ma: int, optional
   :param long_ma: number of data points utilized to calculate long moving average values, defaults to 32
   :type long_ma: int, optional
   :param max_gap: number of minutes a gap between CGM data points can be without having to split the MAGE calculation into multiple segments, defaults to 180
   :type max_gap: int, optional
   :return: the MAGE for the given CGM trace
   :rtype: float
   """
   data = df.reset_index(drop=True)

   config.read('config.ini')
   interval = int(config["variables"]["interval"])

   missing = data[GLUCOSE].isnull()
   # create groups of consecutive missing values
   groups = missing.ne(missing.shift()).cumsum()
   # group by the created groups and count the size of each group, then apply it where values are missing
   size_of_groups = data.groupby([groups, missing])[GLUCOSE].transform('size').where(missing, 0)
   # filter groups where size is greater than 0 and take their indexes
   indexes = size_of_groups[size_of_groups.diff() > (max_gap / interval)].index.tolist()

   if not indexes: # no gaps in data larger than max_gap
      return MAGE_helper(df, short_ma, long_ma)
   else: # calculate MAGE per segment and add them together (weighted)
      indexes.insert(0, 0); indexes.append(None)
      mage = 0
      total_duration = 0
      for i in range(len(indexes) - 1):
         segment = data.iloc[indexes[i]:indexes[i+1]]
         segment = segment.loc[segment[GLUCOSE].first_valid_index():].reset_index(drop=True)
         segment_duration = (segment.iloc[-1][TIME] - segment.iloc[0][TIME]).total_seconds(); total_duration += segment_duration
         mage +=  segment_duration * MAGE_helper(segment, short_ma, long_ma)
      return mage / total_duration

def MAGE_helper(df: pd.DataFrame, short_ma: int = 5, long_ma: int = 32) -> float:
   """Calculates the Mean Amplitude of Glycemic Excursions (MAGE) for a specific segment of a CGM trace.
   Algorithm for calculating MAGE is based on iglu's implementation, and this method is a helper for the MAGE() function.

   :param df: a Pandas DataFrame containing preprocessed CGM data without significant gaps (as defined in the MAGE() function)
   :type df: 'pandas.DataFrame'
   :param short_ma: number of data points utilized to calculate short moving average values, defaults to 5
   :type short_ma: int, optional
   :param long_ma: number of data points utilized to calculate long moving average values, defaults to 32
   :type long_ma: int, optional
   :return: the MAGE for the given segment of a CGM trace
   :rtype: float
   """
   averages = pd.DataFrame()
   averages[GLUCOSE] = df[GLUCOSE]
   averages.reset_index(drop=True, inplace=True)

   if short_ma < 1 or long_ma < 1:
      raise Exception("Moving average spans must be positive, non-zero integers.")

   if short_ma >= long_ma:
      raise Exception("Short moving average span must be smaller than the long moving average span.")
   
   if averages.shape[0] < long_ma:
      return np.nan

   # calculate rolling means, iglu does right align instead of center
   averages["MA_Short"] = averages[GLUCOSE].rolling(window=short_ma, min_periods=1).mean()
   averages["MA_Long"] = averages[GLUCOSE].rolling(window=long_ma, min_periods=1).mean()

   # fill in leading NaNs due to moving average calculation
   averages["MA_Short"].iloc[:short_ma-1] = averages["MA_Short"].iloc[short_ma-1]
   averages["MA_Long"].iloc[:long_ma-1] = averages["MA_Long"].iloc[long_ma-1]
   averages["DELTA_SL"] = averages["MA_Short"] - averages["MA_Long"]
   
   # get crossing points
   glu = lambda i: averages[GLUCOSE].iloc[i]
   average = lambda i: averages["DELTA_SL"].iloc[i]
   crosses_list = [{"location": 0, "type": np.where(average(0) > 0, "peak", "nadir")}]

   for index in range(1, averages.shape[0]):
      current_actual = glu(index)
      current_average = average(index)
      previous_actual = glu(index-1)
      previous_average = average(index-1)

      if not (np.isnan(current_actual) or np.isnan(previous_actual) or np.isnan(current_average) or np.isnan(previous_average)):
         if current_average * previous_average < 0:
            type = np.where(current_average < previous_average, "nadir", "peak")
            crosses_list.append({"location": index, "type": type})   
         elif (not np.isnan(current_average) and (current_average * average(crosses_list[-1]["location"]) < 0)):
            prev_delta = average(crosses_list[-1]["location"])
            type = np.where(current_average < prev_delta, "nadir", "peak")
            crosses_list.append({"location": index, "type": type})

   crosses_list.append({"location": None, "type": np.where(average(-1) > 0, "peak", "nadir")})
   crosses = pd.DataFrame(crosses_list)     

   num_extrema = crosses.shape[0] -  1
   minmax = np.tile(np.nan, num_extrema)
   indexes = pd.Series(np.nan, index=range(num_extrema))

   for index in range(num_extrema):
      s1 = int(np.where(index == 0, crosses["location"].iloc[index], indexes.iloc[index-1]))
      s2 = crosses["location"].iloc[index+1]

      values = averages[GLUCOSE].loc[s1:s2]
      if crosses["type"].iloc[index] == "nadir":
         minmax[index] = np.min(values)
         indexes.iloc[index] = values.idxmin() 
      else:
         minmax[index] = np.max(values)
         indexes.iloc[index] = values.idxmax() 
   
   differences = np.transpose(minmax[:, np.newaxis] - minmax)
   sd = np.std(df[GLUCOSE].dropna())
   N = len(minmax)

   # MAGE+
   mage_plus_heights = []
   mage_plus_tp_pairs = []
   j = 0; prev_j = 0
   while j < N:
      delta = differences[prev_j:j+1,j]

      max_v = np.max(delta)
      i = np.argmax(delta) + prev_j

      if max_v >= sd:
         for k in range(j, N):
            if minmax[k] > minmax[j]:
               j = k
            if (differences[j, k] < (-1 * sd)) or (k == N - 1):
               max_v = minmax[j] - minmax[i]
               mage_plus_heights.append(max_v)
               mage_plus_tp_pairs.append([i, j])

               prev_j = k
               j = k
               break
      else:
         j += 1
   
   # MAGE-
   mage_minus_heights = []
   mage_minus_tp_pairs = [] 
   j = 0; prev_j = 0
   while j < N:
      delta = differences[prev_j:j+1,j]
      min_v = np.min(delta) 
      i = np.argmin(delta) + prev_j

      if min_v <= (-1 * sd):
         for k in range(j, N):
            if minmax[k] < minmax[j]:
               j = k
            if (differences[j, k] > sd) or (k == N - 1):
               min_v = minmax[j] - minmax[i]
               mage_minus_heights.append(min_v)
               mage_minus_tp_pairs.append([i, j, k])

               prev_j = k
               j = k
               break
      else:
         j += 1

   plus_first = len(mage_plus_heights) > 0 and ((len(mage_minus_heights) == 0) or (mage_plus_tp_pairs[0][1] <= mage_minus_tp_pairs[0][0]))
   return float(np.where(plus_first, np.mean(mage_plus_heights), np.mean(np.absolute(mage_minus_heights))))

def m_value(df: pd.DataFrame, r: int = 100) -> float:
   """Calculates the M-value for the given CGM trace.

   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :param r: a reference value utilized for calculating the M-value, defaults to 100
   :type r: int, optional
   :return: the M-value for the given CGM trace
   :rtype: float
   """
   return (abs(10 * np.log10(df[GLUCOSE] / r)) ** 3).mean()

def ROC(df: pd.DataFrame, timedelta: int = 15) -> pd.Series:
   """Returns a Pandas Series with the rate of change in glucose values at every data point

   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :param timedelta: difference in time (in minutes) to utilize when calculating differences between data points, defaults to 15
   :type timedelta: int, optional
   :return: a Pandas Series with the rate of change in glucose values at every data point
   :rtype: 'pandas.Series'
   """
   config.read('config.ini')
   interval = int(config["variables"]["interval"])
   if timedelta < interval:
      raise Exception("Given timedelta must be greater than resampling interval.")

   positiondelta = round(timedelta / interval)
   return df[GLUCOSE].diff(periods=positiondelta) / timedelta

def number_readings(df: pd.DataFrame):
   return df[GLUCOSE].count()

def FBG(df: pd.DataFrame) -> float:
    # Ensure time is in datetime
    df = df.dropna(subset=[GLUCOSE]).copy()
    df['date'] = df[TIME].dt.date

    daily_fbg_means = []
    for day, day_df in df.groupby('date'):
        # Filter data for readings between 6:00 and 7:00 AM
        morning_df = day_df[(day_df[TIME].dt.hour == 6)]
        morning_df = morning_df.sort_values(by=TIME)

        if len(morning_df) >= 6:
            # Take the first 6 readings within 6:00-7:00 AM
            first_6 = morning_df.head(6)
            daily_fbg_means.append(first_6[GLUCOSE].mean())
    
    return np.nan if not daily_fbg_means else np.mean(daily_fbg_means)

def LSBG(df: pd.DataFrame) -> float:
    """Calculates the Lowest Sleeping Blood Glucose (LSBG).
    
    Defined as the mean of the six lowest consecutive glucose measures between
    23:30 and 06:30 (spanning midnight) for each day, averaged over all days.
    
    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the LSBG for the given CGM trace
    :rtype: float
    """
    # Drop rows with missing glucose values
    df = df.dropna(subset=[GLUCOSE]).copy()
    df['date'] = df[TIME].dt.date

    daily_lowest_means = []
    unique_dates = sorted(df['date'].unique())

    # For each date d, nighttime window: d 23:30 -> (d+1) 06:30
    for d in unique_dates:
        start_night = pd.to_datetime(d) + pd.Timedelta(hours=23, minutes=30)
        end_night = pd.to_datetime(d) + pd.Timedelta(days=1, hours=6, minutes=30)

        night_df = df[(df[TIME] >= start_night) & (df[TIME] < end_night)].sort_values(by=TIME)
        
        if len(night_df) >= 6:
            # Compute rolling mean of 6 consecutive readings
            rolling_means = night_df[GLUCOSE].rolling(window=6).mean()
            # Get the minimum of those rolling means for the night
            min_rolling_mean = rolling_means.min()
            if not np.isnan(min_rolling_mean):
                daily_lowest_means.append(min_rolling_mean)

    return np.nan if not daily_lowest_means else np.mean(daily_lowest_means)

def mean_24h(df: pd.DataFrame) -> float:
    """Calculates the Mean 24-h starting from 23:30 to the next day's 23:30 for each day.
    
    For each date d, the 24-hour window is from d 23:30 to (d+1) 23:30.
    We compute the mean glucose within this window for each day, then average across all days.
    
    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the mean 24-h BG value for the given CGM trace
    :rtype: float
    """
    # Drop rows with missing glucose values
    df = df.dropna(subset=[GLUCOSE]).copy()
    df['date'] = df[TIME].dt.date

    daily_means = []
    unique_dates = sorted(df['date'].unique())

    for d in unique_dates:
        start_period = pd.to_datetime(d) + pd.Timedelta(hours=23, minutes=30)
        end_period = start_period + pd.Timedelta(days=1)  # next day's 23:30

        period_df = df[(df[TIME] >= start_period) & (df[TIME] < end_period)]

        if not period_df.empty:
            daily_mean = period_df[GLUCOSE].mean()
            daily_means.append(daily_mean)

    return np.nan if not daily_means else np.mean(daily_means)

def mean_24h_auc(df: pd.DataFrame) -> float:
    """Calculates the mean 24-hour AUC (Area Under the Curve) using the trapezoidal rule,
    with a 24-hour period defined from 23:30 of one day to 23:30 of the next day.
    
    For each date d, the 24-hour period is from d 23:30 to (d+1) 23:30.
    The AUC for that day is the integral of glucose over time.
    We then average over all days to get the mean 24-hour AUC.
    
    Integration is done using scipy.integrate.trapezoid with actual timestamps as 'x'.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the mean 24-hour AUC for the given CGM trace (23:30–23:30)
    :rtype: float
    """
    # Drop rows with missing glucose values
    df = df.dropna(subset=[GLUCOSE]).copy()
    df['date'] = df[TIME].dt.date
    
    daily_aucs = []
    unique_dates = sorted(df['date'].unique())

    for d in unique_dates:
        start_period = pd.to_datetime(d) + pd.Timedelta(hours=23, minutes=30)
        end_period = start_period + pd.Timedelta(days=1)

        day_df = df[(df[TIME] >= start_period) & (df[TIME] < end_period)].sort_values(by=TIME)

        if len(day_df) < 2:
            # Not enough data points to form a meaningful trapezoid
            continue
        
        # Compute time array in hours relative to start_period
        times_in_hours = (day_df[TIME] - start_period).dt.total_seconds() / 3600.0
        glucose_values = day_df[GLUCOSE].values

        # Use scipy's trapezoid to integrate glucose over the 24-hour period
        auc = trapezoid(glucose_values, x=times_in_hours)
        daily_aucs.append(auc)
    
    return np.nan if not daily_aucs else np.mean(daily_aucs)

def mean_daytime(df: pd.DataFrame) -> float:
    """
    Calculates the mean daytime glucose, defined as the mean of all measures 
    between 06:30 and 23:30 for each day, averaged across all days.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the mean daytime glucose level
    :rtype: float
    """
    df = df.dropna(subset=[GLUCOSE]).copy()
    df['date'] = df[TIME].dt.date

    daily_means = []
    for d, day_df in df.groupby('date'):
        start_period = pd.to_datetime(d) + pd.Timedelta(hours=6, minutes=30)
        end_period = pd.to_datetime(d) + pd.Timedelta(hours=23, minutes=30)

        daytime_df = day_df[(day_df[TIME] >= start_period) & (day_df[TIME] < end_period)]

        if not daytime_df.empty:
            daily_mean = daytime_df[GLUCOSE].mean()
            daily_means.append(daily_mean)

    return np.nan if not daily_means else np.mean(daily_means)

def mean_nocturnal(df: pd.DataFrame) -> float:
    """
    Calculates the mean nocturnal glucose, defined as the mean of all measures 
    between 23:30 and 06:30 for each day, averaged across all days.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the mean nocturnal glucose level
    :rtype: float
    """
    df = df.dropna(subset=[GLUCOSE]).copy()
    df['date'] = df[TIME].dt.date

    daily_means = []
    for d, day_df in df.groupby('date'):
        # Define the nighttime window
        start_period = pd.to_datetime(d) + pd.Timedelta(hours=23, minutes=30)
        end_period = pd.to_datetime(d) + pd.Timedelta(days=1, hours=6, minutes=30)

        night_df = day_df[(day_df[TIME] >= start_period) & (day_df[TIME] < end_period)]

        # If no readings in that interval for this particular night, skip it
        if not night_df.empty:
            daily_mean = night_df[GLUCOSE].mean()
            daily_means.append(daily_mean)

    return np.nan if not daily_means else np.mean(daily_means)

def auc_daytime(df: pd.DataFrame) -> float:
    """
    Calculates the mean daytime AUC (Area Under the Curve) of glucose 
    between 06:30 and 23:30 for each day, and then averages these daily AUCs.

    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the mean daytime AUC
    :rtype: float
    """
    # Drop rows with missing glucose values
    df = df.dropna(subset=[GLUCOSE]).copy()
    df['date'] = df[TIME].dt.date

    daily_aucs = []
    for d, day_df in df.groupby('date'):
        start_period = pd.to_datetime(d) + pd.Timedelta(hours=6, minutes=30)
        end_period = pd.to_datetime(d) + pd.Timedelta(hours=23, minutes=30)

        daytime_df = day_df[(day_df[TIME] >= start_period) & (day_df[TIME] < end_period)].sort_values(by=TIME)

        if len(daytime_df) < 2:
            # Not enough data points for integration
            continue

        # Compute time array in hours relative to start_period
        times_in_hours = (daytime_df[TIME] - start_period).dt.total_seconds() / 3600.0
        glucose_values = daytime_df[GLUCOSE].values

        # Use scipy's trapezoid to integrate glucose over the daytime period
        auc = trapezoid(glucose_values, x=times_in_hours)
        daily_aucs.append(auc)

    return np.nan if not daily_aucs else np.mean(daily_aucs)

def nocturnal_auc(df: pd.DataFrame) -> float:
    """
    Calculates the mean nocturnal AUC (Area Under the Curve) of glucose 
    between 23:30 and 06:30 for each day, and then averages these daily AUCs.

    For each date d, we define the nocturnal period as d 23:30 to (d+1) 06:30.
    
    :param df: a Pandas DataFrame containing preprocessed CGM data
    :type df: 'pandas.DataFrame'
    :return: the mean nocturnal AUC
    :rtype: float
    """
    df = df.dropna(subset=[GLUCOSE]).copy()
    df['date'] = df[TIME].dt.date

    daily_aucs = []
    unique_dates = sorted(df['date'].unique())

    for d in unique_dates:
        start_period = pd.to_datetime(d) + pd.Timedelta(hours=23, minutes=30)
        end_period = start_period + pd.Timedelta(hours=6, minutes=30)

        night_df = df[(df[TIME] >= start_period) & (df[TIME] < end_period)].sort_values(by=TIME)

        if len(night_df) < 2:
            # Not enough data points for integration
            continue

        # Compute time array in hours relative to start_period
        times_in_hours = (night_df[TIME] - start_period).dt.total_seconds() / 3600.0
        glucose_values = night_df[GLUCOSE].values

        # Use scipy's trapezoid to integrate glucose over the nocturnal period
        auc = trapezoid(glucose_values, x=times_in_hours)
        daily_aucs.append(auc)

    return np.nan if not daily_aucs else np.mean(daily_aucs)


def compute_features(id: str, data: pd.DataFrame) -> dict[str, any]:
   """Calculates statistics and metrics for a single patient within the given DataFrame

   :param id: the patient to calculate features for
   :type id: str
   :param data: Pandas DataFrame containing preprocessed CGM data for one or more patients
   :type data: 'pandas.DataFrame'
   :return: a dictionary (with each key referring to the name of a statistic or metric)
   :rtype: dict[str, any]
   """
   summary = summary_stats(data)

   features = {
      ID: id,
      "ADRR": ADRR(data),
      "COGI": COGI(data),
      "CONGA": CONGA(data),
      "CV": CV(data),
      "Daytime AUC": auc_daytime(data),
      "eA1c": eA1c(data),
      "FBG": FBG(data),
      "First Quartile": summary[1],
      "GMI": GMI(data),
      "GRADE": GRADE(data),
      "GRADE (euglycemic)": GRADE_eugly(data),
      "GRADE (hyperglycemic)": GRADE_hyper(data),
      "GRADE (hypoglycemic)": GRADE_hypo(data),
      "GRI": GRI(data),
      "GVP": GVP(data),
      "HBGI": HBGI(data),
      "Hyperglycemia Index": hyper_index(data),
      "Hypoglycemia Index": hypo_index(data),
      "IGC": IGC(data),
      "J-Index": j_index(data),
      "LBGI": LBGI(data),
      "LSBG": LSBG(data),
      "MAG": MAG(data),
      "MAGE": MAGE(data),
      "Maximum": summary[4],
      "Mean": mean(data),
      "Mean 24h Glucose": mean_24h(data),
      "Mean 24h AUC": mean_24h_auc(data),
      "Mean Absolute Differences": mean_absolute_differences(data),
      "Mean Daytime": mean_daytime(data),
      "Mean Nocturnal": mean_nocturnal(data),
      "Median": summary[2],
      "Median Absolute Deviation": median_absolute_deviation(data),
      "Minimum": summary[0],
      "MODD": MODD(data),
      "M-Value": m_value(data),
      "Nocturnal AUC": nocturnal_auc(data),
      "Number of Readings": number_readings(data),
      "Percent Time Above Range (180)": percent_time_above_range(data),
      "Percent Time Below Range (70)": percent_time_below_range(data),
      "Percent Time in Hyperglycemia": percent_time_in_hyperglycemia(data),
      "Percent Time in Hyperglycemia (level 0)": percent_time_in_level_0_hyperglycemia(data),
      "Percent Time in Hyperglycemia (level 1)": percent_time_in_level_1_hyperglycemia(data),
      "Percent Time in Hyperglycemia (level 2)": percent_time_in_level_2_hyperglycemia(data),
      "Percent Time in Hypoglycemia": percent_time_in_hypoglycemia(data),
      "Percent Time in Hypoglycemia (level 1)": percent_time_in_level_1_hypoglycemia(data),
      "Percent Time in Hypoglycemia (level 2)": percent_time_in_level_2_hypoglycemia(data),
      "Percent Time In Range (70-180)": percent_time_in_range(data),
      "Percent Time In Tight Range (70-140)": percent_time_in_tight_range(data),
      "SD": SD(data),
      "Third Quartile": summary[3],
   }
   return features


def create_features(dataset: pd.DataFrame) -> pd.DataFrame:
   """Takes in a multiindexed Pandas DataFrame containing CGM data for multiple patients/datasets, and
   returns a single indexed Pandas DataFrame containing summary metrics in the form of one row per patient/dataset

   :param dataset: a Pandas DataFrame containing the CGM data to calculate metrics for
   :type dataset: 'pandas.DataFrame'
   :return: a Pandas DataFrame with each row representing a patient in 'dataset' and each column representing a specific statistic or metric
   :rtype: 'pandas.DataFrame'
   """
   with Pool() as pool:
      features = pool.starmap(compute_features, dataset.groupby(ID))
   features = pd.DataFrame(features).set_index([ID])
   return features



import pandas as pd
import numpy as np
from scipy.integrate import trapezoid
import configparser
import glob, os, zipfile, tempfile
import math

dir_path = os.path.dirname(os.path.realpath(__file__))
config_path = os.path.join(dir_path, "config.ini")
config = configparser.ConfigParser()
config.read(config_path)
ID = config['variables']['id']
GLUCOSE = config['variables']['glucose']
TIME = config['variables']['time']
BEFORE = config['variables']['before']
AFTER = config['variables']['after']
TYPE = config['variables']['type']
DESCRIPTION = config['variables']['description']

def import_events(
   path: str, 
   id: str,
   name: str = None,
   day_col: str = "Day",
   time_col: str = "Time",
   before: int = 60,
   after: int = 60,
   type: str = "imported event"
) -> pd.DataFrame:
   """Bulk imports events from standalone .csv files or from those within a given directory or .zip file
   
   :param path: the path of the directory/zip/csv to import from
   :type path: str
   :param id: the identification of the patient that the imported events belong to
   :type id: str
   :param day_col: the name of the column specifying the day the event occurred (year, month, and specific day), defaults to 'Day'
   :type day_col: str, optional
   :param time_col: the name of the column specifying what time during the day the event occurred, defaults to 'Time'
   :type time_col: str, optional
   :param before: the amount of minutes to also look at before the event timestamp, defaults to 60
   :type before: int, optional
   :param after: the amount of minutes to also look at after the event timestamp, defaults to 60
   :type after: int, optional
   :param type: the type of event to classify all the imported events as, defaults to 'imported event'
   :type type: str, optional
   :return: a Pandas DataFrame containing all the imported events
   :rtype: 'pandas.DataFrame'
   """
   ext = os.path.splitext(path)[1]

   # path leads to directory
   if ext == "":
      if not os.path.isdir(path):
         raise ValueError("Directory does not exist")
      else:
         return import_events_directory(path, id, day_col, time_col, before, after, type)
   
   # check if path leads to .zip or .csv
   if ext.lower() in [".csv", ".zip"]:
      if not os.path.isfile(path):
         raise ValueError("File does not exist")
   else:
      raise ValueError("Invalid file type")

   # path leads to .csv
   if ext.lower() == ".csv":
      return import_events_csv(path, id, day_col, time_col, before, after, type)

   # otherwise has to be a .zip file
   with zipfile.ZipFile(path, 'r') as zip_ref:
   # create a temporary directory to pull from
      with tempfile.TemporaryDirectory() as temp_dir:
         zip_ref.extractall(temp_dir)
         dir = name or path.split("/")[-1].split(".")[0]
         return import_events_directory((temp_dir + "/" + dir), id, day_col, time_col, before, after, type)

def import_events_directory(
   path: str,
   id: str,
   day_col: str = "Day",
   time_col: str = "Time",
   before: int = 60,
   after: int = 60,
   type: str = "imported event"
) -> pd.DataFrame:
    """Bulk imports events from .csv files within a given directory
   
   :param path: the path of the directory to import from
   :type path: str
   :param id: the identification of the patient that the imported events belong to
   :type id: str
   :param day_col: the name of the column specifying the day the event occurred (year, month, and specific day), defaults to 'Day'
   :type day_col: str, optional
   :param time_col: the name of the column specifying what time during the day the event occurred, defaults to 'Time'
   :type time_col: str, optional
   :param before: the amount of minutes to also look at before the event timestamp, defaults to 60
   :type before: int, optional
   :param after: the amount of minutes to also look at after the event timestamp, defaults to 60
   :type after: int, optional
   :param type: the type of event to classify all the imported events as, defaults to 'imported event'
   :type type: str, optional
   :return: a Pandas DataFrame containing all the imported events
   :rtype: 'pandas.DataFrame'
   """
    csv_files = glob.glob(path + "/*.csv")

    if len(csv_files) == 0:
       raise Exception("No CSV files found.")

    return pd.concat(import_events_csv(file, id, day_col, time_col, before, after, type) for file in csv_files)

def import_events_csv(
   path: str,
   id: str,
   day_col: str = "Day",
   time_col: str = "Time",
   before: int = 60,
   after: int = 60,
   type: str = "imported event"
) -> pd.DataFrame:
   """Bulk imports events from a single .csv file
   
   :param path: the path of the .csv file to import from
   :type path: str
   :param id: the identification of the patient that the imported events belong to
   :type id: str
   :param day_col: the name of the column specifying the day the event occurred (year, month, and specific day), defaults to 'Day'
   :type day_col: str, optional
   :param time_col: the name of the column specifying what time during the day the event occurred, defaults to 'Time'
   :type time_col: str, optional
   :param before: the amount of minutes to also look at before the event timestamp, defaults to 60
   :type before: int, optional
   :param after: the amount of minutes to also look at after the event timestamp, defaults to 60
   :type after: int, optional
   :param type: the type of event to classify all the imported events as, defaults to 'imported event'
   :type type: str, optional
   :return: a Pandas DataFrame containing all the imported events
   :rtype: 'pandas.DataFrame'
   """
   df = pd.read_csv(path)
   csv_name = os.path.splitext(path)[0]

   events = pd.DataFrame()
   events[TIME] = pd.to_datetime(df[day_col] + " " + df[time_col])
   events[BEFORE] = before
   events[AFTER] = after
   events[TYPE] = type
   events[DESCRIPTION] = df["Food Name"] if "Food Name" in df.columns else ("imported event #" + (events.index + 1).astype(str) + f"from {csv_name}")
   events.insert(0, ID, id)

   return events.dropna(subset=[TIME])

def _episodes_helper(
   df: pd.DataFrame, 
   id: str, 
   type: str, 
   threshold: int, 
   level: int, 
   min_length: int, 
   end_length: int
) -> pd.DataFrame:
   """Retrieves all episodes of a specific type/level for a specific patient within the given CGM data

   :param df: Pandas DataFrame containing preprocessed CGM data
   :type df: pandas.DataFrame
   :param id: identification of the patient to retrieve episodes for
   :type id: str
   :param type: type of episode ('hypo' or 'hyper')
   :type type: str
   :param threshold: threshold (in mg/dL) above/below which glucose values are considered as part of an episode
   :type threshold: int
   :param level: the level the retrieved episodes are (0, 1, or 2)
   :type level: int
   :param min_length: minimum duration (in minutes) required for excursions, defaults to 15
   :type min_length: int, optional
   :param end_length: minimum amount of time (in minutes) that the glucose values must be within typical ranges 
      at the end of an excursion, defaults to 15
   :type end_length: int, optional
   :return: a Pandas DataFrame containing all episodes of a specific type/level for a specific patient within the given CGM data
   :rtype: pandas.DataFrame
   """

   config.read('config.ini')
   interval = int(config["variables"]["interval"])
   timegap = lambda timedelta: timedelta.total_seconds() / 60
   episodes = pd.DataFrame()

   data = df.copy(); data.reset_index(drop=True, inplace=True)
   episode_df = df[(df[GLUCOSE] <= threshold)].copy() if type == "hypo" else df[df[GLUCOSE] >= threshold].copy()
   episode_df.reset_index(drop=True, inplace=True)
   episode_df["gap"] = episode_df[TIME].diff().apply(timegap)

   edges = episode_df.index[episode_df["gap"] != interval].to_list()
   edges.append(-1)

   get = lambda loc, col: episode_df.iloc[loc][col]
   index = 0
   while index < len(edges) - 1:
      offset = 0 if (index == len(edges) - 2) else 1
      end_i = edges[index + 1] - offset # index of the end of the episode (inclusive! - that's what the offset is for)
      start_i = edges[index] # index of the start of the episode
      start_time = get(start_i, TIME)
      end_time = get(end_i, TIME)
      episode_length = timegap(end_time - start_time)

      if episode_length >= min_length: # check if episode lasts longer than 15 min
         if offset != 0: # not the very last episode
            end_counts = math.ceil(end_length / interval)
            
            end_index = data.index[data[TIME] == end_time].to_list()[0]
            end_data = data.iloc[end_index + 1 : end_index + 1 + end_counts][GLUCOSE]
            outside_threshold = np.where(end_data >= threshold, True, False) if type == "hypo" else np.where(end_data <= threshold, True, False)
            if False in outside_threshold: # check if episode ends within 15 min
               edges.pop(index + 1) # this episode does not end within 15 min, so combine this episode with the next
               continue

         description = f"{start_time} to {end_time} level {level} {type}glycemic episode"
         event = pd.DataFrame.from_records([{ID: id, TIME: start_time, BEFORE: 0, AFTER: episode_length, 
                                             TYPE: f"{type} level {level} episode", DESCRIPTION: description}])
         episodes = pd.concat([episodes, event]) 
      
      index += 1

   return episodes

def get_episodes(
    df: pd.DataFrame,
    hypo_lvl2: int = 54,
    hypo_lvl1: int = 70,
    hyper_lvl0: int = 140,
    hyper_lvl1: int = 180,
    hyper_lvl2: int = 250,
    min_length: int = 15,
    end_length: int = 15
) -> pd.DataFrame:
   """Retrieves all episodes within the given CGM data

   :param df: Pandas DataFrame containing preprocessed CGM data
   :type df: pandas.DataFrame
   :param hypo_lvl2: threshold (in mg/dL) below which glucose values are considered level 2 hypoglycemic, defaults to 54
   :type hypo_lvl2: int, optional
   :param hypo_lvl1: threshold (in mg/dL) below which glucose values are considered level 1 hypoglycemic, defaults to 70
   :type hypo_lvl1: int, optional
   :param hyper_lvl0: threshold (in mg/dL) above which glucose values are considered level 0 hyperglycemic, defaults to 140
   :type hyper_lvl0: int, optional
   :param hyper_lvl1: threshold (in mg/dL) above which glucose values are considered level 1 hyperglycemic, defaults to 180
   :type hyper_lvl1: int, optional
   :param hyper_lvl2: threshold (in mg/dL) above which glucose values are considered level 2 hyperglycemic, defaults to 250
   :type hyper_lvl2: int, optional
   :param min_length: minimum duration (in minutes) required for excursions, defaults to 15
   :type min_length: int, optional
   :param end_length: minimum amount of time (in minutes) that the glucose values must be within typical ranges 
      at the end of an excursion, defaults to 15
   :type end_length: int, optional
   :return: a Pandas DataFrame containing all episodes within the given CGM data
   :rtype: pandas.DataFrame
   """
   output = pd.DataFrame()
   for id, data in df.groupby(ID):
      episodes = pd.concat([_episodes_helper(data, id, "hyper", hyper_lvl0, 0, min_length, end_length),
                            _episodes_helper(data, id, "hyper", hyper_lvl1, 1, min_length, end_length),
                            _episodes_helper(data, id, "hyper", hyper_lvl2, 2, min_length, end_length),
                            _episodes_helper(data, id, "hypo", hypo_lvl1, 1, min_length, end_length),
                            _episodes_helper(data, id, "hypo", hypo_lvl2, 2, min_length, end_length)])
      
      episodes.sort_values(by=[TIME], inplace=True)
      output = pd.concat([output, episodes])

   return output 

def get_excursions(
   df: pd.DataFrame, 
   z: int = 2, 
   min_length: int = 15,
   end_length: int = 15
) -> pd.DataFrame:
   """Retrieves all excursions within the given CGM data

   :param df: Pandas DataFrame containing preprocessed CGM data
   :type df: pandas.DataFrame
   :param z: the number of standard deviations away from the mean that should define an 'excursion', defaults to 2
   :type z: int, optional
   :param min_length: minimum duration (in minutes) required for excursions, defaults to 15
   :type min_length: int, optional
   :param end_length: minimum amount of time (in minutes) that the glucose values must be within typical ranges 
      at the end of an excursion, defaults to 15
   :type end_length: int, optional
   :return: a Pandas DataFrame containing all excursions within the given CGM data
   :rtype: pandas.DataFrame
   """

   excursions = pd.DataFrame()

   config.read('config.ini')
   interval = int(config["variables"]["interval"])
   for id, data in df.groupby(ID):
      data.reset_index(drop=True, inplace=True)
      sd = data[GLUCOSE].std()
      mean = data[GLUCOSE].mean()
      upper = mean + (z * sd)
      lower = mean - (z * sd)

      peaks = data[(data[GLUCOSE].shift(1) < data[GLUCOSE]) & (data[GLUCOSE].shift(-1) < data[GLUCOSE])][TIME].copy()
      peaks.reset_index(drop=True, inplace=True)
      nadirs = data[(data[GLUCOSE].shift(1) > data[GLUCOSE]) & (data[GLUCOSE].shift(-1) > data[GLUCOSE])][TIME].copy()
      nadirs.reset_index(drop=True, inplace=True)

      outliers = data[(data[GLUCOSE] >= upper) | (data[GLUCOSE] <= lower)].copy()
      outliers.reset_index(drop=True, inplace=True)

      # calculate the differences between each of the timestamps
      timegap = lambda timedelta: timedelta.total_seconds() / 60
      outliers["gaps"] = outliers[TIME].diff().apply(timegap)

      edges = outliers.index[outliers["gaps"] != interval].to_list()
      edges.append(-1)
      i = 0
      while i < len(edges) - 1:
         type = "hyper" if outliers.iloc[edges[i]][GLUCOSE] > mean else "hypo"
         offset = 0 if i == len(edges) - 2 else 1
         start_time = outliers.iloc[edges[i]][TIME]
         start_index = data.index[data[TIME] == start_time].to_list()[0]
         end_time = outliers.iloc[edges[i+1] - offset][TIME]
         end_index = data.index[data[TIME] == end_time].to_list()[0]

         excursion_length = timegap(end_time - start_time)
         if excursion_length >= min_length:
            if offset != 0: # not the very last episode
               end_counts = math.ceil(end_length / interval)
               
               last_index = data.reset_index().index[data[TIME] == end_time].to_list()[0]
               last_data = data.iloc[last_index + 1 : last_index + 1 + end_counts][GLUCOSE]
               outside_threshold = np.where(last_data <= upper if type == "hyper" else last_data >= lower, True, False)
               if False in outside_threshold: # check if excursion ends within 15 min
                  edges.pop(i + 1) # this excursion does not end within 15 min, so combine this episode with the next
                  continue
               
            outliers.set_index(TIME, inplace=True)
            last_point = edges[i+1] if offset != 0 else None
            timestamp = outliers.iloc[edges[i]:last_point][GLUCOSE].idxmax() if type == "hyper" else outliers.iloc[edges[i]:last_point][GLUCOSE].idxmin()
            outliers.reset_index(inplace=True)

            extrema = peaks if type == "hypo" else nadirs
            if start_index != 0:
               if not extrema[extrema <= start_time].empty: start_time = extrema[extrema <= start_time].iloc[-1]
            if end_index != data.shape[0] - 1:
               if not extrema[extrema >= end_time].empty: end_time = extrema[extrema >= end_time].iloc[0]
            
            description = f"{start_time} to {end_time} {type}glycemic excursion"
            event = pd.DataFrame.from_records([{ID: id, TIME: timestamp, BEFORE: timegap(timestamp - start_time), 
                                                AFTER: timegap(end_time - timestamp), 
                                                TYPE: f"{type} excursion", DESCRIPTION: description}])
            excursions = pd.concat([excursions, event])
         
         i += 1

   return excursions

def get_curated_events(df: pd.DataFrame) -> pd.DataFrame:
   """Retrieves all curated events (episodes and excursions) for all the patients within the given DataFrame

   :param df: a Pandas DataFrame containing preprocessed CGM data
   :type df: 'pandas.DataFrame'
   :return: a Pandas DataFrame (in the usual event structure defined by the package) containing all curated events for all the patients within the given DataFrame
   :rtype: 'pandas.DataFrame'
   """
   return pd.concat([get_episodes(df), get_excursions(df)])

def retrieve_event_data(
    df: pd.DataFrame,
    events: pd.DataFrame,
) -> pd.DataFrame:
    """Returns a multiindexed Pandas DataFrame containing only patient data during the respective given events
    :param df: a Pandas DataFrame containing the preprocessed CGM traces to retrieve event subsets from
    :type df: 'pandas.DataFrame'
    :param events: a single indexed Pandas DataFrame, with each row specifying a single event in the form of
                   an id, a datetime, # of hours before the datetime to include, # of hours after to include, and a description
    :type events: 'pandas.DataFrame'
    :return: a multi-indexed Pandas DataFrame, with each index referring to a subset of CGM trace that was found within 'df' and occurs during a single event within 'events' 
    :rtype: 'pandas.DataFrame'
    """
    event_data = pd.DataFrame()
    for index, row in events.to_frame().T.iterrows():
      id = row[ID]
      if id in df.index:
         datetime = pd.Timestamp(row[TIME])
         initial = datetime - pd.Timedelta(row[BEFORE], "m")
         final = datetime + pd.Timedelta(row[AFTER], "m")

         patient_data = df.loc[id]
         data = patient_data[(patient_data[TIME] >= initial) & (patient_data[TIME] <= final)].copy()

         data[ID] = id
         data[DESCRIPTION] = row[DESCRIPTION]

         event_data = pd.concat([event_data, data])

    #if event_data.shape[0] != 0:
      #event_data = event_data.set_index(["id"])

    return event_data

def event_summary(events: pd.DataFrame) -> pd.Series:
   """Returns the number of events per unique event type found within 'events'

   :param events: a Pandas DataFrame containing events (as per package guidelines)
   :type events: 'pandas.DataFrame'
   :return: a Pandas Series containing the number of events per unique event type found within 'events'
   :rtype: 'pandas.Series'
   """
   return events[TYPE].value_counts()

def AUC(df: pd.DataFrame) -> float:
    """Calculates the total Area-Under-Curve (AUC) for the given CGM trace

    :param df: a Pandas DataFrame containing the CGM trace to calculate the AUC of
    :type df: 'pandas.DataFrame'
    :return: the AUC of the given CGM trace
    :rtype: float
    """
    config.read('config.ini')
    interval = int(config["variables"]["interval"])
    return trapezoid(df[GLUCOSE], dx=interval)

def iAUC(df: pd.DataFrame, level: float) -> float:
    """Calculates the incremental Area-Under-Curve (iAUC) for the given CGM trace

    :param df: a Pandas DataFrame containing the CGM trace to calculate the AUC of
    :type df: 'pandas.DataFrame'
    :param level: the threshold above which to calculate iAUC
    :type level: float
    :return: the iAUC of the given CGM trace
    :rtype: float
    """
    data = df.copy()
    data[GLUCOSE] = abs(data[GLUCOSE] - level)
    data.loc[data[GLUCOSE] < 0, GLUCOSE] = 0
    return AUC(data)

def baseline(df: pd.DataFrame) -> float:
    """Returns the baseline glucose level for the given CGM trace

    :param df: a Pandas DataFrame containing the CGM trace to retrieve the baseline glucose level for
    :type df: 'pandas.DataFrame'
    :return: the baseline glucose level of the given CGM trace
    :rtype: float
    """
    return df[GLUCOSE].iloc[0]

def peak(df: pd.DataFrame) -> float:
    """Returns the maximum glucose level for the given CGM trace

    :param df: a Pandas DataFrame containing the CGM trace to retrieve the maximum glucose level for
    :type df: 'pandas.DataFrame'
    :return: the maximum glucose level of the given CGM trace
    :rtype: float
    """
    return np.max(df[GLUCOSE])

def nadir(df: pd.DataFrame) -> float:
   """Returns the minimum glucose level for the given CGM trace

    :param df: a Pandas DataFrame containing the CGM trace to retrieve the minimum glucose level for
    :type df: 'pandas.DataFrame'
    :return: the minimum glucose level of the given CGM trace
    :rtype: float
    """
   return np.min(df[GLUCOSE])

def delta(df: pd.DataFrame) -> float:
    """Returns the difference in maximum and baseline glucose levels (delta) for the given CGM trace

    :param df: a Pandas DataFrame containing the CGM trace to retrieve the delta for
    :type df: 'pandas.DataFrame'
    :return: the delta of the given CGM trace
    :rtype: float
    """
    return abs(peak(df) - baseline(df))

def post_event_glucoses(data: pd.DataFrame, event_time: pd.Timestamp, times: list[int], glucose_col: str = GLUCOSE) -> dict:
    """
    Returns the glucose values closest to the specified times (in minutes) after the given event_time.

    :param data: Pandas DataFrame containing the CGM data
    :type data: pd.DataFrame
    :param event_time: The time of the event
    :type event_time: pd.Timestamp
    :param times: A list of integers representing the number of minutes after event_time for which to find the glucose values
    :type times: list[int]
    :param glucose_col: The name of the glucose column in the data, defaults to GLUCOSE
    :type glucose_col: str, optional
    :return: A dictionary where keys are strings like "X-min Post Event" and values are the corresponding glucose readings or np.nan if not found
    :rtype: dict
    """
    result = {}

    # Always include 0-min to have a reference point
    if 0 not in times:
       times = [0] + times

    for t in times:
        key = f"{t}-min Post Event"
        result[key] = np.nan
        post_time = event_time + pd.Timedelta(minutes=t)

        # Check if the desired time is within the range of the data
        if not data.empty and data[TIME].min() <= post_time <= data[TIME].max():
            closest_idx = (data[TIME] - post_time).abs().idxmin()
            result[key] = data.loc[closest_idx, glucose_col]

    return result

def post_event_aucs(data: pd.DataFrame, event_time: pd.Timestamp, durations: list[int], glucose_col: str = GLUCOSE) -> dict:
    """
    Calculates AUC values for multiple durations (in minutes) starting from the given event_time.
    
    :param data: Pandas DataFrame containing the CGM data
    :type data: pd.DataFrame
    :param event_time: The time of the event
    :type event_time: pd.Timestamp
    :param durations: A list of integers representing the number of minutes after event_time for which to calculate the AUC
    :type durations: list[int]
    :param glucose_col: The name of the glucose column in the data, defaults to GLUCOSE
    :type glucose_col: str, optional
    :return: A dictionary where keys are strings like "X-min AUC" and values are the corresponding AUC readings or np.nan if no data is available
    :rtype: dict
    """
    result = {}
    for d in durations:
        key = f"{d}-min AUC"
        end_time = event_time + pd.Timedelta(minutes=d)
        subset = data[(data[TIME] >= event_time) & (data[TIME] <= end_time)].copy()

        if subset.empty:
            result[key] = np.nan
        else:
            result[key] = AUC(subset)
    return result

def event_metrics(
      df: pd.DataFrame,
      event: pd.Series,
      post_times: list[int] = [60, 120],
      post_auc_times: list[int] = [120]
) -> pd.DataFrame:
   """Calculates basic metrics for events (baseline, peak, delta, iAUC, and
   0-h, 1-h, and 2-h post event glucose values, and 2-h post event AUC)

   :param df: Pandas DataFrame containing preprocessed CGM data
   :type df: pandas.DataFrame
   :param event: Pandas Series with fields that represent an 'event'
   :type event: pandas.Series
   :param post_times: A list of integers representing the number of minutes after event_time for which to find the glucose values
   :type post_times: list[int], optional (defaults to [0, 60, 120] for 0-h, 1-h and 2-h post event)
   :return: Pandas DataFrame containing the basic metrics for the given event
   :rtype: pandas.DataFrame
   """
   id = event[ID]

   datetime = pd.Timestamp(event[TIME])
   initial = datetime - pd.Timedelta(event[BEFORE], "m")
   final = datetime + pd.Timedelta(event[AFTER], "m")

   patient_data = df.loc[id]
   data = patient_data[(patient_data[TIME] >= initial) & (patient_data[TIME] <= final)].copy()

   metrics = pd.Series()
   metrics["Baseline"] = baseline(data)
   metrics["Peak"] = peak(data)
   metrics["Delta"] = delta(data)
   metrics["AUC"] = AUC(data)
   metrics["iAUC"] = iAUC(data, baseline(data))

   # Get post-event glucose values (including 0-min)
   post_values = post_event_glucoses(data, datetime, post_times, GLUCOSE)
   for k, v in post_values.items():
      metrics[k] = v
   
   # Compute deltas from 0-min Post Event
   zero_min_val = metrics["0-min Post Event"]
   for t in post_times:
      if t == 0:
         continue
      post_key = f"{t}-min Post Event"
      delta_key = f"{t}-min Delta"
      if post_key in metrics and not np.isnan(metrics[post_key]) and not np.isnan(zero_min_val):
         metrics[delta_key] = metrics[post_key] - zero_min_val
      else:
         metrics[delta_key] = np.nan

   auc_values = post_event_aucs(data, datetime, post_auc_times, GLUCOSE)
   for k, v in auc_values.items():
      metrics[k] = v

   return metrics.to_frame().T

def create_event_features(
    df: pd.DataFrame,
    events: pd.DataFrame,
) -> pd.DataFrame:
   """Returns a multi-indexed Pandas DataFrame containing metrics for the patient data during their respective 'events'
   
   :param df: a Pandas DataFrame containing all the relevant patient CGM data to generate event metrics for
   :type df: 'pandas.Series'
   :param events: a single indexed Pandas DataFrame, with each row specifying a single event in the form of
                  an id, a datetime, # of hours before the datetime to include, # of hours after to include, and a desc
   :type events: 'pandas.DataFrame'
   :return: a multi-indexed Pandas DataFrame containing metrics for the patient data during their respective 'events'
   """
   event_features = {}
   for id in df.index.unique():
      sub_features = {}
      for type, sub_events in events[events[ID] == id].groupby(TYPE):
         sub_features.update(create_event_features_helper(df.loc[id], sub_events, type))
      event_features[id] = sub_features

   return pd.DataFrame(event_features).T

def create_event_features_helper(
   df: pd.DataFrame,
   sub_events: pd.DataFrame,
   type: str,     
) -> dict[str, float]:
   """Calculates aggregate event-based metrics for a single patient and type of event. Helper method for 'create_event_features()'.

   :param df: Pandas DataFrame containing the CGM trace for a single patient
   :type df: 'pandas.DataFrame'
   :param sub_events: Pandas DataFrame containing events of only one type solely for the patient whose CGM trace is also given
   :type sub_events: 'pandas.DataFrame'
   :param type: the type of event that 'sub_events' contains
   :type type: str
   :return: a dictionary with str-type keys that refer to the name of the calculated features and float-type values
   :rtype: dict[str, float]
   """
   
   features = {
      f"Mean {type} Duration": [],
      f"Mean Glucose During {type}s": [],
      f"Mean Upwards Slope of {type}s (mg/dL per min)": [],
      f"Mean Downwards Slope of {type}s (mg/dL per min)": [],
      f"Mean Minimum Glucose of {type}s": [],
      f"Mean Maximum Glucose of {type}s": [],
      f"Mean Amplitude of {type}s": [],
      f"Mean iAUC of {type}s": []
   }
   
   for _, event in sub_events.iterrows():
      event_data = retrieve_event_data(df, event)

      duration = event[AFTER] - event[BEFORE]
      features[f"Mean {type} Duration"].append(duration)

      features[f"Mean Glucose During {type}s"].append(event_data[GLUCOSE].mean())
      features[f"Mean Minimum Glucose of {type}s"] = nadir(event_data)
      features[f"Mean Maximum Glucose of {type}s"] = peak(event_data)

      event_time = event[TIME]
      closest_idx = (event_data[TIME] - event_time).abs().idxmin()
      event_glucose = event_data.loc[closest_idx, GLUCOSE]

      peak_glucose = peak(event_data)
      peak_time = event_data.loc[event_data[GLUCOSE].idxmax(), TIME]
      amplitude = peak_glucose - event_glucose
      features[f"Mean Amplitude of {type}s"].append(abs(amplitude))

      time_diff_to_peak = (peak_time - event_time).total_seconds() / 60.0
      slope_to_peak = (peak_glucose - event_glucose) / time_diff_to_peak if time_diff_to_peak != 0 else np.nan
      features[f"Mean Upwards Slope of {type}s (mg/dL per min)"].append(slope_to_peak)

      end_time = event_data[TIME].iloc[-1]
      end_glucose = event_data[GLUCOSE].iloc[-1]
      time_diff_peak_to_end = (end_time - peak_time).total_seconds() / 60.0
      slope_peak_to_end = (end_glucose - peak_glucose) / time_diff_peak_to_end if time_diff_peak_to_end != 0 else np.nan
      features[f"Mean Downwards Slope of {type}s (mg/dL per min)"].append(slope_peak_to_end)

      features[f"Mean iAUC of {type}s"].append(iAUC(event_data, event_glucose))

   features = {k: np.mean(v) for k, v in features.items()}
   features[f"Mean # of {type}s per day"] = sub_events.shape[0] / len(df[TIME].dt.date.unique())
   return features